from pslab.instrument.logic_analyzer import LogicAnalyzer
from pslab.instrument.multimeter import Multimeter
from pslab.instrument.oscilloscope import Oscilloscope
from pslab.instrument.power_supply import PowerSupply
from pslab.instrument.waveform_generator import PWMGenerator, WaveformGenerator
from pslab.sciencelab import ScienceLab



"""Control the PSLab's I2C bus and devices connected on the bus.

Examples
--------
Set I2C bus speed to 400 kbit/s:

>>> from pslab.bus.i2c import I2CMaster, I2CSlave
>>> bus = I2CMaster()
>>> bus.configure(frequency=4e5)

Scan for connected devices:

>>> bus.scan()
[96, 104]

Connect to the PSLab's built-in DS1307 RTC:

>>> rtc = I2CSlave(address=104)
"""
import logging
from typing import List

import pslab.protocol as CP
from pslab.serial_handler import SerialHandler
from pslab.external.sensorlist import sensors

logger = logging.getLogger(__name__)


class I2CMaster:
    """I2C bus controller.

    Handles slave independent functionality with the I2C port.

    Parameters
    ----------
    device : :class:`SerialHandler`, optional
        Serial connection to PSLab device. If not provided, a new one will be
        created.
    """

    _MIN_BRGVAL = 2
    _MAX_BRGVAL = 511
    # Specs say typical delay is 110 ns to 130 ns; 150 ns from testing.
    _SCL_DELAY = 150e-9

    def __init__(self, device: SerialHandler = None):
        self._device = device if device is not None else SerialHandler()
        self._device.send_byte(CP.I2C_HEADER)
        self._device.send_byte(CP.I2C_INIT)
        self._device.get_ack()
        self.configure(125e3)  # 125 kHz is as low as the PSLab can go.

    def configure(self, frequency: float):
        """Configure bus frequency.

        Parameters
        ----------
        frequency : float
            Frequency of SCL in Hz.
        """
        brgval = int((1 / frequency - self._SCL_DELAY) * CP.CLOCK_RATE - 2)

        if self._MIN_BRGVAL <= brgval <= self._MAX_BRGVAL:
            self._device.send_byte(CP.I2C_HEADER)
            self._device.send_byte(CP.I2C_CONFIG)
            self._device.send_int(brgval)
            self._device.get_ack()
        else:
            min_frequency = self._get_i2c_frequency(self._MAX_BRGVAL)
            max_frequency = self._get_i2c_frequency(self._MIN_BRGVAL)
            e = f"Frequency must be between {min_frequency} and {max_frequency} Hz."
            raise ValueError(e)

    def _get_i2c_frequency(self, brgval: int) -> float:
        return 1 / ((brgval + 2) / CP.CLOCK_RATE + self._SCL_DELAY)

    def scan(self) -> List[int]:
        """Scan I2C port for connected devices.

        Returns
        -------
        addrs : list of int
            List of 7-bit addresses on which slave devices replied.
        """
        addrs = []

        for address in range(1, 128):  # 0 is the general call address.
            slave = I2CSlave(address, self._device)

            if slave.ping():
                addrs.append(address)
                logger.info(
                    f"Response from slave on {hex(address)} "
                    + f"({sensors.get(address, 'None')})."
                )

        return addrs

    @property
    def _status(self):
        """int: Contents of the I2C2STAT register.

        bit 15  ACKSTAT: Acknowledge Status
                1 = NACK received from slave.
                0 = ACK received from slave.
        bit 14  TRSTAT: Transmit Status
                1 = Master transmit is in progress (8 bits + ACK).
                0 = Master transmit is not in progress.
        bit 10  BCL: Bus Collision Detect
                1 = A bus collision has been detected.
                0 = No collision.
        bit 7   IWCOL: I2C Write Collision Detect
                1 = An attempt to write to the I2C2TRN register failed because
                    the I2C module is busy.
                0 = No collision.
        bit 6   I2COV: I2C Receive Overflow Flag
                1 = A byte is received while the I2C2RCV register is still
                    holding the previous byte.
                0 = No overflow.
        bit 4   P: Stop
                1 = Indicates that a Stop bit has been detected last.
                0 = Stop bit was not detected last.
        bit 3   S: Start
                1 = Indicates that a Start (or Repeated Start) bit has been
                    detected last.
                0 = Start bit was not detected last.
        bit 1   RBF: Receive Buffer Full Status
                1 = Receive completes; the I2C2RCV register is full.
                0 = Receive is not complete; the I2C2RCV register is empty.
        bit 0   TBF: Transmit Buffer Full Status
                1 = Transmit is in progress; the I2C2TRN register is full.
                0 = Transmit completes; the I2C2TRN register is empty.
        """
        self._device.send_byte(CP.I2C_HEADER)
        self._device.send_byte(CP.I2C_STATUS)
        status = self._device.get_int()
        self._device.get_ack()
        return status


class I2CSlave:
    """I2C slave device.

    Parameters
    ----------
    address : int
        7-bit I2C device address.
    device : :class:`SerialHandler`, optional
        Serial interface for communicating with the PSLab device. If not
        provided, a new one will be created.

    Attributes
    ----------
    address : int
        7-bit I2C device address.
    """

    _ACK = 0
    _READ = 1
    _WRITE = 0

    def __init__(
        self,
        address: int,
        device: SerialHandler = None,
    ):
        self._device = device if device is not None else SerialHandler()
        self.address = address
        self._running = False
        self._mode = None

    def ping(self) -> bool:
        """Ping the I2C slave.

        Returns
        -------
        response : bool
            True is slave responded, False otherwise.
        """
        response = self._start(self._READ)
        self._stop()
        return response == self._ACK

    def _start(self, mode: int) -> int:
        """Initiate I2C transfer.

        Parameters
        ----------
        mode : {0, 1}
            0: write
            1: read

        Returns
        -------
        response : int
            Response from I2C slave device.
        """
        if self._mode == mode:
            return self._ACK

        self._device.send_byte(CP.I2C_HEADER)
        secondary = CP.I2C_START if not self._running else CP.I2C_RESTART
        self._device.send_byte(secondary)
        self._device.send_byte((self.address << 1) | mode)
        response = self._device.get_ack() >> 4  # ACKSTAT
        self._running = True
        return response

    def _stop(self):
        """Stop I2C transfer."""
        if self._running or self._mode:
            self._device.send_byte(CP.I2C_HEADER)
            self._device.send_byte(CP.I2C_STOP)
            self._device.get_ack()
            self._running = False
            self._mode = None

    def read(self, bytes_to_read: int, register_address: int = 0x0) -> bytearray:
        """Read data from I2C device.

        This method relies on the slave device auto incrementing its internal
        pointer after each read. Most devices do, but it is not part of the
        I2C standard. Refer to the device's documentation.

        If the slave device does not auto increment, use one of
        :meth:`read_byte`, :meth:`read_int`, or :meth:`read_long` instead,
        depending on how wide the device registers are.

        Parameters
        ----------
        bytes_to_read : int
            Number of bytes to read from slave device.
        register_address : int, optional
            Slave device internal memory address to read from. The default
            value is 0x0.

        Returns
        -------
        data : bytearray
            Read data as a bytearray.
        """
        self._device.send_byte(CP.I2C_HEADER)
        self._device.send_byte(CP.I2C_READ_BULK)
        self._device.send_byte(self.address)
        self._device.send_byte(register_address)
        self.pointer = register_address
        self._device.send_byte(bytes_to_read)
        data = self._device.read(bytes_to_read)
        self._device.get_ack()
        return bytearray(data)

    def read_byte(self, register_address: int = 0x0) -> int:
        """Read a single byte from the I2C slave.

        Parameters
        ----------
        register_address : int, optional
            Slave device internal memory address to read from. The default
            value is 0x0.

        Returns
        -------
        data : int
            A byte interpreted as a uint8.
        """
        return self.read(1, register_address)[0]

    def read_int(self, register_address: int = 0x0) -> int:
        """Read a two byte value from the I2C slave.

        Parameters
        ----------
        register_address : int, optional
            Slave device internal memory address to read from. The default
            value is 0x0.

        Returns
        -------
        data : int
            Two bytes interpreted as a uint16.
        """
        data = self.read(2, register_address)
        return CP.ShortInt.unpack(data)[0]

    def read_long(self, register_address: int = 0x0) -> int:
        """Read a four byte value from the I2C slave.

        Parameters
        ----------
        register_address : int, optional
            Slave device internal memory address to read from. The default
            value is 0x0.

        Returns
        -------
        data : int
            Four bytes interpreted as a uint32.
        """
        data = self.read(4, register_address)
        return CP.Integer.unpack(data)[0]

    def write(self, bytes_to_write: bytearray, register_address: int = 0x0):
        """Write data to I2C slave.

        This method relies on the slave device auto incrementing its internal
        pointer after each write. Most devices do, but it is not part of the
        I2C standard. Refer to the device's documentation.

        If the slave device does not auto increment, use one of
        :meth:`write_byte`, :meth:`write_int`, or :meth:`write_long` instead,
        depending on how wide the device registers are.

        Parameters
        ----------
        bytes_to_write : bytearray
            Data to write to the slave.
        register_address : int
            Slave device internal memory address to write to. The default
            value is 0x0.
        """
        self._device.send_byte(CP.I2C_HEADER)
        self._device.send_byte(CP.I2C_WRITE_BULK)
        self._device.send_byte(self.address)
        self._device.send_byte(len(bytes_to_write) + 1)  # +1 is register address.
        self._device.send_byte(register_address)

        for byte in bytes_to_write:
            self._device.send_byte(byte)

        self._device.get_ack()

    def write_byte(self, data: int, register_address: int = 0x0):
        """Write a single byte to the I2C slave.

        Parameters
        ----------
        data : int
            An integer that fits in a uint8.
        register_address : int, optional
            Slave device internal memory address to write to. The default
            value is 0x0.
        """
        self.write(CP.Byte.pack(data), register_address)

    def write_int(self, data: int, register_address: int = 0x0):
        """Write a two byte value to the I2C slave.

        Parameters
        ----------
        data : int
            An integer that fits in a uint16.
        register_address : int, optional
            Slave device internal memory address to write to. The default
            value is 0x0.
        """
        self.write(CP.ShortInt.pack(data), register_address)

    def write_long(self, data: int, register_address: int = 0x0):
        """Write a four byte value to the I2C slave.

        Parameters
        ----------
        data : int
            An integer that fits in a uint32.
        register_address : int, optional
            Slave device internal memory address to write to. The default
            value is 0x0.
        """
        self.write(CP.Integer.pack(data), register_address)



import pslab.protocol as CP


class HCSR04:
    """Read data from ultrasonic distance sensor HC-SR04/HC-SR05.

    Sensors must have separate trigger and output pins.
    First a 10 µs pulse is output on SQ1. SQ1 must be connected to the TRIG pin
    on the sensor prior to use.

    Upon receiving this pulse, the sensor emits a sequence of sound pulses, and
    the logic level of its output pin (which we will monitor via LA1) is also
    set high.  The logic level goes LOW when the sound packet returns to the
    sensor, or when a timeout occurs.

    The ultrasound sensor outputs a series of eight sound pulses at 40 kHz,
    which corresponds to a time period of 25 µs per pulse. These pulses reflect
    off of the nearest object in front of the sensor, and return to it. The
    time between sending and receiving of the pulse packet is used to estimate
    the distance. If the reflecting object is either too far away or absorbs
    sound, less than eight pulses may be received, and this can cause a
    measurement error of 25 µs which corresponds to 8 mm.

    The sensor requires a 5 V supply. You may set SQ2 to HIGH:
    >>> psl.pwm_generator.set_state(sq2=True)
    and use that as the power supply.

    Returns 0 in case of timeout.
    """

    def __init__(self, device):
        self._device = device

    def estimate_distance(self):
        self._device.send_byte(CP.NONSTANDARD_IO)
        self._device.send_byte(CP.HCSR04_HEADER)

        timeout_msb = int((0.3 * 64e6)) >> 16
        self._device.send_int(timeout_msb)

        A = self._device.get_long()
        B = self._device.get_long()
        tmt = self._device.get_int()
        self._device.get_ack()

        if (tmt >= timeout_msb or B == 0):
            return 0

        return 330 * (B - A + 20) / 64e6 / 2



"""Classes and functions related to the PSLab's logic analyzer instrument.

Example
-------
>>> from pslab import LogicAnalyzer
>>> la = LogicAnalyzer()
>>> t = la.capture(channels=2, events=1600, modes=["falling", "any"])
"""

import time
from collections import OrderedDict
from typing import Dict, List, Tuple, Union

import numpy as np

import pslab.protocol as CP
from pslab.instrument.digital import DigitalInput, DIGITAL_INPUTS, MODES
from pslab.serial_handler import ADCBufferMixin, SerialHandler


class LogicAnalyzer(ADCBufferMixin):
    """Investigate digital signals on up to four channels simultaneously.

    Parameters
    ----------
    device : :class:`SerialHandler`, optional
        Serial connection to PSLab device. If not provided, a new one will be
        created.

    Attributes
    ----------
    trigger_channel : str
        See :meth:`configure_trigger`.
    trigger_mode : str
        See :meth:`configure_trigger`.
    """

    _PRESCALERS = {
        0: 1,
        1: 8,
        2: 64,
        3: 256,
    }

    # When capturing multiple channels, there is a two clock cycle
    # delay between channels.
    _CAPTURE_DELAY = 2

    def __init__(self, device: SerialHandler = None):
        self._device = SerialHandler() if device is None else device
        self._channels = {d: DigitalInput(d) for d in DIGITAL_INPUTS}
        self.trigger_channel = "LA1"
        self._trigger_channel = self._channels["LA1"]
        self.trigger_mode = "disabled"
        self._trigger_mode = 0
        self._prescaler = 0
        self._channel_one_map = "LA1"
        self._channel_two_map = "LA2"
        self._trimmed = 0

    def measure_frequency(
        self, channel: str, simultaneous_oscilloscope: bool = False, timeout: float = 1
    ) -> float:
        """Measure the frequency of a signal.

        Parameters
        ----------
        channel : {"LA1", "LA2", "LA3", "LA4"}
            Name of the digital input channel in which to measure the
            frequency.
        simultaneous_oscilloscope: bool, optional
            Set this to True if you need to use the oscilloscope at the same
            time. Uses firmware instead of software to measure the frequency,
            which may fail and return 0. Will not give accurate results above
            10 MHz. The default value is False.
        timeout : float, optional
            Timeout in seconds before cancelling measurement. The default value
            is 1 second.

        Returns
        -------
        frequency : float
            The signal's frequency in Hz.
        """
        if simultaneous_oscilloscope:
            return self._measure_frequency_firmware(channel, timeout)
        else:
            tmp = self._channel_one_map
            self._channel_one_map = channel
            t = self.capture(1, 2, modes=["sixteen rising"], timeout=timeout)[0]
            self._channel_one_map = tmp

            try:
                period = (t[1] - t[0]) * 1e-6 / 16
                frequency = period ** -1
            except IndexError:
                frequency = 0

            if frequency >= 1e7:
                frequency = self._get_high_frequency(channel)

            return frequency

    def _measure_frequency_firmware(
        self, channel: str, timeout: float, retry: bool = True
    ) -> float:
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.GET_FREQUENCY)
        self._device.send_int(int(timeout * 64e6) >> 16)
        self._device.send_byte(self._channels[channel].number)
        self._device.wait_for_data(timeout)

        error = self._device.get_byte()
        t = [self._device.get_long() for a in range(2)]
        self._device.get_ack()
        edges = 16
        period = (t[1] - t[0]) / edges / CP.CLOCK_RATE

        if error or period == 0:
            # Retry once.
            if retry:
                return self._measure_frequency_firmware(channel, timeout, False)
            else:
                return 0
        else:
            return period ** -1

    def _get_high_frequency(self, channel: str) -> float:
        """Measure high frequency signals using firmware.

        The input frequency is fed to a 32 bit counter for a period of 100 ms.
        The value of the counter at the end of 100 ms is used to calculate the
        frequency.
        """
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.GET_ALTERNATE_HIGH_FREQUENCY)
        self._device.send_byte(self._channels[channel].number)
        scale = self._device.get_byte()
        counter_value = self._device.get_long()
        self._device.get_ack()

        return scale * counter_value / 1e-1  # 100 ms sampling

    def measure_interval(
        self, channels: List[str], modes: List[str], timeout: float = 1
    ) -> float:
        """Measure the time between two events.

        This method cannot be used at the same time as the oscilloscope.

        Parameters
        ----------
        channels : List[str]
            A pair of digital inputs, LA1, LA2, LA3, or LA4. Both can be the
            same.
        modes : List[str]
            Type of logic event to listen for on each channel. See
            :class:`DigitalInput` for available modes.
        timeout : float, optional
            Timeout in seconds before cancelling measurement. The default value
            is 1 second.

        Returns
        -------
        interval : float
            Time between events in microseconds. A negative value means that
            the event on the second channel happend first.
        """
        tmp_trigger = self._trigger_channel.name
        self.configure_trigger(channels[0], self.trigger_mode)
        tmp_map = self._channel_one_map, self._channel_two_map
        self._channel_one_map = channels[0]
        self._channel_two_map = channels[1]

        if channels[0] == channels[1]:
            # 34 edges contains 17 rising edges, i.e two
            # 'every sixteenth rising edge' events.
            t = self.capture(1, 34, modes=["any"], timeout=timeout)[0]
            initial = self.get_initial_states()[self._channel_one_map]
            t1 = self._get_first_event(t, modes[0], initial)

            if modes[0] == modes[1]:
                idx = 1 if modes[1] == "any" else 2
                initial = initial if idx == 2 else not initial
                t2 = self._get_first_event(t[idx:], modes[1], initial)
            else:
                t2 = self._get_first_event(t, modes[1], initial)
        else:
            t1, t2 = self.capture(2, 1, modes=modes, timeout=timeout)

            t1, t2 = t1[0], t2[0]

        self.configure_trigger(tmp_trigger, self.trigger_mode)
        self._channel_one_map = tmp_map[0]
        self._channel_two_map = tmp_map[1]

        return t2 - t1

    @staticmethod
    def _get_first_event(events: np.ndarray, mode: str, initial: bool) -> np.ndarray:
        if mode == "any":
            return events[0]
        elif mode == "rising":
            return events[int(initial)]
        elif mode == "falling":
            return events[int(not initial)]
        elif mode == "four rising":
            return events[initial::2][3]
        elif mode == "sixteen rising":
            return events[initial::2][15]

    def measure_duty_cycle(self, channel: str, timeout: float = 1) -> Tuple[float]:
        """Measure duty cycle and wavelength.

        This method cannot be used at the same time as the oscilloscope.

        Parameters
        ----------
        channel : {"LA1", "LA2", "LA3", "LA4"}
            Digital input on which to measure.
        timeout : float, optional
            Timeout in seconds before cancelling measurement. The default value
            is 1 second.

        Returns
        -------
        wavelength : float
            Wavelength in microseconds.
        duty_cycle : float
             Duty cycle as a value between 0 - 1.
        """
        tmp_trigger_mode = self.trigger_mode
        tmp_trigger_channel = self._trigger_channel.name
        self.configure_trigger(trigger_channel=channel, trigger_mode="rising")
        tmp_map = self._channel_one_map
        self._channel_one_map = channel
        t = self.capture(1, 3, modes=["any"], timeout=timeout)[0]
        self._channel_one_map = tmp_map
        self.configure_trigger(tmp_trigger_channel, tmp_trigger_mode)

        period = t[2] - t[0]
        # First change is HIGH -> LOW since we trigger on rising.
        duty_cycle = 1 - (t[1] - t[0]) / period

        return period, duty_cycle

    def capture(
        self,
        channels: Union[int, str, List[str]],
        events: int = CP.MAX_SAMPLES // 4,
        timeout: float = 1,
        modes: List[str] = 4 * ("any",),
        e2e_time: float = None,
        block: bool = True,
    ) -> Union[List[np.ndarray], None]:
        """Capture logic events.

        This method cannot be used at the same time as the oscilloscope.

        Parameters
        ----------
        channels : {1, 2, 3, 4} or str or list of str
            Number of channels to capture events on. Events will be captured on
            LA1, LA2, LA3, and LA4, in that order. Alternatively, the name of
            of a single digital input, or a list of two names of digital inputs
            can be provided. In that case, events will be captured only on that
            or those specific channels.
        events : int, optional
            Number of logic events to capture on each channel. The default and
            maximum value is 2500.
        timeout : float, optional
            Timeout in seconds before cancelling measurement in blocking mode.
            If the timeout is reached, the events captured up to that point
            will be returned. The default value is 1 second.
        modes : List[str], optional
            List of strings specifying the type of logic level change to
            capture on each channel. See :class:`DigitalInput` for available
            modes. The default value is ("any", "any", "any", "any").
        e2e_time : float, optional
            The maximum time between events in seconds. This is only required
            in three and four channel mode, which uses 16-bit counters as
            opposed to 32-bit counters which are used in one and two channel
            mode. The 16-bit counter normally rolls over after 1024 µs, so if
            the time between events is greater than that the timestamp
            calculations will be incorrect. By setting this to a value greater
            than 1024 µs, the counter will be slowed down by a prescaler, which
            can extend the maximum allowed event-to-event time gap to up to
            262 ms. If the time gap is greater than that, three and four
            channel mode cannot be used. One and two channel mode supports
            timegaps up to 67 seconds.
        block : bool, optional
            Whether to block while waiting for events to be captured. If this
            is False, this method will return None immediately and the captured
            events must be manually fetched by calling :meth:`fetch_data`. The
            default value is True.

        Returns
        -------
        events : list of numpy.ndarray or None
            List of numpy.ndarrays containing timestamps in microseconds when
            logic events were detected, or None if block is False. The length
            of the list is equal to the number of channels that were used to
            capture events, and each list element corresponds to a channel.

        Raises
        ------
        ValueError if too many events are requested, or
        ValueError if too many channels are selected.
        """
        channels = self._check_arguments(channels, events)
        self.stop()
        self._prescaler = 0
        self.clear_buffer(CP.MAX_SAMPLES)
        self._invalidate_buffer()
        self._configure_trigger(channels)
        modes = [MODES[m] for m in modes]
        start_time = time.time()

        for e, c in enumerate(
            [self._channel_one_map, self._channel_two_map, "LA3", "LA4"][:channels]
        ):
            c = self._channels[c]
            c.events_in_buffer = events
            c.datatype = "long" if channels < 3 else "int"
            c.buffer_idx = 2500 * e * (1 if c.datatype == "int" else 2)
            c._logic_mode = modes[e]

        if channels == 1:
            self._capture_one()
        elif channels == 2:
            self._capture_two()
        else:
            self._capture_four(e2e_time)

        if block:
            # Discard 4:th channel if user asked for 3.
            timestamps = self.fetch_data()[:channels]
            progress = min([len(t) for t in timestamps])
            while progress < events:
                timestamps = self.fetch_data()[:channels]
                progress = min([len(t) for t in timestamps])
                if time.time() - start_time >= timeout:
                    break
                if progress >= CP.MAX_SAMPLES // 4 - self._trimmed:
                    break
        else:
            return

        for e, t in enumerate(timestamps):
            timestamps[e] = t[:events]  # Don't surprise the user with extra events.

        return timestamps

    def _check_arguments(self, channels: Union[int, str, List[str]], events: int):
        if isinstance(channels, str):
            self._channel_one_map = channels
            channels = 1

        if isinstance(channels, list):
            self._channel_one_map = channels[0]
            self._channel_two_map = channels[1]
            channels = 2

        max_events = CP.MAX_SAMPLES // 4

        if events > max_events:
            raise ValueError(f"Events must be fewer than {max_events}.")
        elif channels < 0 or channels > 4:
            raise ValueError("Channels must be between 1-4.")

        return channels

    def _capture_one(self):
        self._channels[self._channel_one_map]._prescaler = 0
        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.START_ALTERNATE_ONE_CHAN_LA)
        self._device.send_int(CP.MAX_SAMPLES // 4)
        self._device.send_byte(
            (self._channels[self._channel_one_map].number << 4)
            | self._channels[self._channel_one_map]._logic_mode
        )
        self._device.send_byte(
            (self._channels[self._channel_one_map].number << 4) | self._trigger_mode
        )
        self._device.get_ack()

    def _capture_two(self):
        for c in list(self._channels.values())[:2]:
            c._prescaler = 0

        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.START_TWO_CHAN_LA)
        self._device.send_int(CP.MAX_SAMPLES // 4)
        self._device.send_byte((self._trigger_channel.number << 4) | self._trigger_mode)
        self._device.send_byte(
            self._channels[self._channel_one_map]._logic_mode
            | (self._channels[self._channel_two_map]._logic_mode << 4)
        )
        self._device.send_byte(
            self._channels[self._channel_one_map].number
            | (self._channels[self._channel_two_map].number << 4)
        )
        self._device.get_ack()

    def _capture_four(self, e2e_time: float):
        rollover_time = (2 ** 16 - 1) / CP.CLOCK_RATE
        e2e_time = 0 if e2e_time is None else e2e_time

        if e2e_time > rollover_time * self._PRESCALERS[3]:
            raise ValueError("Timegap too big for four channel mode.")
        elif e2e_time > rollover_time * self._PRESCALERS[2]:
            self._prescaler = 3
        elif e2e_time > rollover_time * self._PRESCALERS[1]:
            self._prescaler = 2
        elif e2e_time > rollover_time:
            self._prescaler = 1
        else:
            self._prescaler = 0

        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.START_FOUR_CHAN_LA)
        self._device.send_int(CP.MAX_SAMPLES // 4)
        self._device.send_int(
            self._channels["LA1"]._logic_mode
            | (self._channels["LA2"]._logic_mode << 4)
            | (self._channels["LA3"]._logic_mode << 8)
            | (self._channels["LA4"]._logic_mode << 12)
        )
        self._device.send_byte(self._prescaler)

        try:
            trigger = {
                0: 4,
                1: 8,
                2: 16,
            }[self._trigger_channel.number] | self._trigger_mode
        except KeyError:
            e = "Triggering is only possible on LA1, LA2, or LA3."
            raise NotImplementedError(e)

        self._device.send_byte(trigger)
        self._device.get_ack()

    def fetch_data(self) -> List[np.ndarray]:
        """Collect captured logic events.

        It is possible to call fetch_data while the capture routine is still running.
        Doing so will not interrupt the capture process. In multi-channel mode, the
        number of timestamps may differ between channels when fetch_data is called
        before the capture is complete.

        Returns
        -------
        data : list of numpy.ndarray
            List of numpy.ndarrays holding timestamps in microseconds when logic events
            were detected. The length of the list is equal to the number of channels
            that were used to capture events, and each list element corresponds to a
            channel.
        """
        counter_values = []
        channels = list(
            OrderedDict.fromkeys(
                [self._channel_one_map, self._channel_two_map, "LA3", "LA4"]
            )
        )
        for c in channels:
            c = self._channels[c]

            if c.events_in_buffer:
                if c.datatype == "long":
                    raw_timestamps = self._fetch_long(c)
                else:
                    raw_timestamps = self._fetch_int(c)
                counter_values.append(raw_timestamps)

        prescaler = [1 / 64, 1 / 8, 1.0, 4.0][self._prescaler]

        timestamps = []
        capture_delay = self._CAPTURE_DELAY if self._prescaler == 0 else 0
        for e, cv in enumerate(counter_values):
            adjusted_counter = cv + e * capture_delay
            timestamps.append(adjusted_counter * prescaler)

        return timestamps

    def _fetch_long(self, channel: DigitalInput) -> np.ndarray:
        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.FETCH_LONG_DMA_DATA)
        self._device.send_int(CP.MAX_SAMPLES // 4)
        self._device.send_byte(channel.buffer_idx // 5000)
        raw = self._device.read(CP.MAX_SAMPLES)
        self._device.get_ack()

        raw_timestamps = [
            CP.Integer.unpack(raw[a * 4 : a * 4 + 4])[0]
            for a in range(CP.MAX_SAMPLES // 4)
        ]
        raw_timestamps = np.array(raw_timestamps)
        raw_timestamps = np.trim_zeros(raw_timestamps, "b")
        pretrim = len(raw_timestamps)
        raw_timestamps = np.trim_zeros(raw_timestamps, "f")
        self._trimmed = pretrim - len(raw_timestamps)

        return raw_timestamps

    def _fetch_int(self, channel: DigitalInput) -> np.ndarray:
        raw_timestamps = self.fetch_buffer(CP.MAX_SAMPLES // 4, channel.buffer_idx)
        raw_timestamps = np.array(raw_timestamps)
        raw_timestamps = np.trim_zeros(raw_timestamps, "b")
        pretrim = len(raw_timestamps)
        raw_timestamps = np.trim_zeros(raw_timestamps, "f")
        self._trimmed = pretrim - len(raw_timestamps)

        for i, diff in enumerate(np.diff(raw_timestamps)):
            if diff <= 0:  # Counter has rolled over.
                raw_timestamps[i + 1 :] += 2 ** 16 - 1

        return raw_timestamps

    def get_progress(self) -> int:
        """Return the number of captured events per channel held in the buffer.

        Returns
        -------
        progress : int
            Number of events held in buffer. If multiple channels have events
            in buffer, the lowest value will be returned.
        """
        active_channels = []
        a = 0
        for c in self._channels.values():
            if c.events_in_buffer:
                active_channels.append(a * (1 if c.datatype == "int" else 2))
                a += 1

        p = CP.MAX_SAMPLES // 4
        progress = self._get_initial_states_and_progress()[1]
        for a in active_channels:
            p = min(progress[a], p)

        return p

    def get_initial_states(self) -> Dict[str, bool]:
        """Return the initial state of each digital input at the beginning of capture.

        Returns
        -------
        dict of four str: bool pairs
            Dictionary containing pairs of channel names and the corresponding initial
            state, e.g. {'LA1': True, 'LA2': True, 'LA3': True, 'LA4': False}.
            True means HIGH, False means LOW.
        """
        return self._get_initial_states_and_progress()[0]

    def get_xy(
        self, timestamps: List[np.ndarray], initial_states: Dict[str, bool] = None
    ) -> List[np.ndarray]:
        """Turn timestamps into plottable data.

        Parameters
        ----------
        timestamps : list of numpy.ndarray
            List of timestamps as returned by :meth:`capture` or
            :meth:`fetch_data`.
        initial_states : dict of str, bool
            Initial states of digital inputs at beginning of capture, as
            returned by :meth:`get_initial_states`. If no additional capture
            calls have been issued before calling :meth:`get_xy`, this can be
            omitted.

        Returns
        -------
        list of numpy.ndarray
            List of x, y pairs suitable for plotting using, for example,
            matplotlib.pyplot.plot. One pair of x and y values is returned for
            each list of timestamps given as input.
        """
        xy = []
        initial_states = (
            initial_states if initial_states is not None else self.get_initial_states()
        )

        for e, c in enumerate(
            [self._channel_one_map, self._channel_two_map, "LA3", "LA4"][
                : len(timestamps)
            ]
        ):
            c = self._channels[c]
            if c.events_in_buffer:
                x, y = c._get_xy(initial_states[c.name], timestamps[e])
                xy.append(x)
                xy.append(y)

        return xy

    def _get_initial_states_and_progress(self) -> Tuple[Dict[str, bool], List[int]]:
        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.GET_INITIAL_DIGITAL_STATES)
        initial = self._device.get_int()
        progress = [0, 0, 0, 0]
        progress[0] = (self._device.get_int() - initial) // 2
        progress[1] = (self._device.get_int() - initial) // 2 - CP.MAX_SAMPLES // 4
        progress[2] = (self._device.get_int() - initial) // 2 - 2 * CP.MAX_SAMPLES // 4
        progress[3] = (self._device.get_int() - initial) // 2 - 3 * CP.MAX_SAMPLES // 4
        s = self._device.get_byte()
        initial_states = {
            "LA1": (s & 1 != 0),
            "LA2": (s & 2 != 0),
            "LA3": (s & 4 != 0),
            "LA4": (s & 8 != 0),
        }
        self._device.get_byte()  # INITIAL_DIGITAL_STATES_ERR
        self._device.get_ack()

        for e, i in enumerate(progress):
            if i == 0:
                progress[e] = CP.MAX_SAMPLES // 4
            elif i < 0:
                progress[e] = 0

        return initial_states, progress

    def configure_trigger(self, trigger_channel: str, trigger_mode: str):
        """Set up trigger channel and trigger condition.

        Parameters
        ----------
        trigger_channel : {"LA1", "LA2", "LA3", "LA4"}
            The digital input on which to trigger.
        trigger_mode : {"disabled", "falling", "rising"}
            The type of logic level change on which to trigger.
        """
        self.trigger_channel = trigger_channel
        self._trigger_channel = self._channels[trigger_channel]
        self.trigger_mode = trigger_mode

    def _configure_trigger(self, channels: int):
        # For some reason firmware uses different values for trigger_mode
        # depending on number of channels.
        if channels == 1:
            self._trigger_mode = {
                "disabled": 0,
                "any": 1,
                "falling": 2,
                "rising": 3,
                "four rising": 4,
                "sixteen rising": 5,
            }[self.trigger_mode]
        elif channels == 2:
            self._trigger_mode = {
                "disabled": 0,
                "falling": 3,
                "rising": 1,
            }[self.trigger_mode]
        elif channels == 4:
            self._trigger_mode = {
                "disabled": 0,
                "falling": 1,
                "rising": 3,
            }[self.trigger_mode]

    def stop(self):
        """Stop a running :meth:`capture` function."""
        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.STOP_LA)
        self._device.get_ack()

    def get_states(self) -> Dict[str, bool]:
        """Return the current state of the digital inputs.

        Returns
        -------
        dict of four str: bool pairs
            Dictionary containing pairs of channel names and the corresponding
            current state, e.g. {'LA1': True, 'LA2': True, 'LA3': True,
            'LA4': False}. True means HIGH, False means LOW.
        """
        self._device.send_byte(CP.DIN)
        self._device.send_byte(CP.GET_STATES)
        s = self._device.get_byte()
        self._device.get_ack()
        return {
            "LA1": (s & 1 != 0),
            "LA2": (s & 2 != 0),
            "LA3": (s & 4 != 0),
            "LA4": (s & 8 != 0),
        }

    def count_pulses(
        self, channel: str = "FRQ", interval: float = 1, block: bool = True
    ) -> Union[int, None]:
        """Count pulses on a digital input.

        The counter is 16 bits, so it will roll over after 65535 pulses. This
        method can be used at the same time as the oscilloscope.

        Parameters
        ----------
        channel : {"LA1", "LA2", "LA3", "LA4", "FRQ"}, optional
            Digital input on which to count pulses. The default value is "FRQ".
        interval : float, optional
            Time in seconds during which to count pulses. The default value is
            1 second.
        block : bool, optional
            Whether to block while counting pulses or not. If False, this
            method will return None, and the pulses must be manually fetched
            using :meth:`fetch_pulse_count`. Additionally, the interval
            argument has no meaning if block is False; the counter will keep
            counting even after the interval time has expired. The default
            value is True.

        Returns
        -------
        Union[int, None]
            Number of pulses counted during the interval, or None if block is
            False.
        """
        self._reset_prescaler()
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.START_COUNTING)
        self._device.send_byte(self._channels[channel].number)
        self._device.get_ack()

        if block:
            time.sleep(interval)
        else:
            return

        return self.fetch_pulse_count()

    def fetch_pulse_count(self) -> int:
        """Return the number of pulses counted since calling :meth:`count_pulses`.

        Returns
        -------
        int
            Number of pulses counted since calling :meth:`count_pulses`.
        """
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.FETCH_COUNT)
        count = self._device.get_int()
        self._device.get_ack()
        return count

    def _reset_prescaler(self):
        self._device.send_byte(CP.TIMING)
        self._device.send_byte(CP.START_FOUR_CHAN_LA)
        self._device.send_int(0)
        self._device.send_int(0)
        self._device.send_byte(0)
        self._device.send_byte(0)
        self._device.get_ack()
        self.stop()
        self._prescaler = 0

    def _invalidate_buffer(self):
        for c in self._channels.values():
            c.events_in_buffer = 0
            c.buffer_idx = None



from __future__ import print_function


def connect(route, **args):
    return MLX90614(route, **args)


class MLX90614():
    NUMPLOTS = 1
    PLOTNAMES = ['Temp']
    ADDRESS = 0x5A
    name = 'PIR temperature'

    def __init__(self, I2C, **args):
        self.I2C = I2C
        self.ADDRESS = args.get('address', self.ADDRESS)
        self.OBJADDR = 0x07
        self.AMBADDR = 0x06

        self.source = self.OBJADDR

        self.name = 'Passive IR temperature sensor'
        self.params = {'readReg': {'dataType': 'integer', 'min': 0, 'max': 0x20, 'prefix': 'Addr: '},
                       'select_source': ['object temperature', 'ambient temperature']}

        try:
            print('switching baud to 100k')
            self.I2C.configI2C(100e3)
        except Exception as e:
            print('FAILED TO CHANGE BAUD RATE', e.message)

    def select_source(self, source):
        if source == 'object temperature':
            self.source = self.OBJADDR
        elif source == 'ambient temperature':
            self.source = self.AMBADDR

    def readReg(self, addr):
        x = self.getVals(addr, 2)
        print(hex(addr), hex(x[0] | (x[1] << 8)))

    def getVals(self, addr, numbytes):
        vals = self.I2C.readBulk(self.ADDRESS, addr, numbytes)
        return vals

    def getRaw(self):
        vals = self.getVals(self.source, 3)
        if vals:
            if len(vals) == 3:
                return [((((vals[1] & 0x007f) << 8) + vals[0]) * 0.02) - 0.01 - 273.15]
            else:
                return False
        else:
            return False

    def getObjectTemperature(self):
        self.source = self.OBJADDR
        val = self.getRaw()
        if val:
            return val[0]
        else:
            return False

    def getAmbientTemperature(self):
        self.source = self.AMBADDR
        val = self.getRaw()
        if val:
            return val[0]
        else:
            return False



"""Objects related to the PSLab's analog input channels.

This module contains several module level variables with details on the analog
inputs' capabilities, including possible gain values, voltage ranges, and
firmware-interal enumeration.

This module also contains the AnalogInput class, an instance of which functions
as a model of a particular analog input.
"""

import logging
from typing import List, Union

import numpy as np

logger = logging.getLogger(__name__)

GAIN_VALUES = (1, 2, 4, 5, 8, 10, 16, 32)

ANALOG_CHANNELS = (
    "CH1",
    "CH2",
    "CH3",
    "MIC",
    "CAP",
    "RES",
    "VOL",
    "AN4",
)

INPUT_RANGES = {
    "CH1": (16.5, -16.5),  # Specify inverted channels explicitly by reversing range!
    "CH2": (16.5, -16.5),
    "CH3": (-3.3, 3.3),  # external gain control analog input
    "MIC": (-3.3, 3.3),  # connected to MIC amplifier
    "CAP": (0, 3.3),
    "RES": (0, 3.3),
    "VOL": (0, 3.3),
    "AN4": (0, 3.3),
}

PIC_ADC_MULTIPLEX = {
    "CH1": 3,
    "CH2": 0,
    "CH3": 1,
    "MIC": 2,
    "AN4": 4,
    "RES": 7,
    "CAP": 5,
    "VOL": 8,
}


class AnalogInput:
    """Model of the PSLab's analog inputs, used to scale raw values to voltages.

    Parameters
    ----------
    name : {'CH1', 'CH2', 'CH3', 'MIC', 'CAP', 'RES', 'VOL'}
        Name of the analog channel to model.

    Attributes
    ----------
    samples_in_buffer : int
        Number of samples collected on this channel currently being held in the
        device's ADC buffer.
    buffer_idx : int or None
        Location in the device's ADC buffer where the samples are stored. None
        if no samples captured by this channel are currently held in the
        buffer.
    chosa : int
        Number used to refer to this channel in the firmware.
    """

    def __init__(self, name: str):
        self._name = name
        self._resolution = 2 ** 10 - 1

        if self._name == "CH1":
            self.programmable_gain_amplifier = 1
            self._gain = 1
        elif self._name == "CH2":
            self.programmable_gain_amplifier = 2
            self._gain = 1
        else:
            self.programmable_gain_amplifier = None
            self._gain = 1

        self.samples_in_buffer = 0
        self.buffer_idx = None
        self._scale = np.poly1d(0)
        self._unscale = np.poly1d(0)
        self.chosa = PIC_ADC_MULTIPLEX[self._name]
        self._calibrate()

    @property
    def gain(self) -> Union[int, None]:
        """int: Analog gain.

        Setting a new gain level will automatically recalibrate the channel.
        On channels other than CH1 and CH2 gain is None.

        Raises
        ------
        TypeError
            If gain is set on a channel which does not support it.
        ValueError
            If a gain value other than 1, 2, 4, 5, 8, 10, 16, 32 is set.
        """
        if self._name in ("CH1", "CH2"):
            return self._gain
        else:
            return None

    @gain.setter
    def gain(self, value: Union[int, float]):
        if self._name not in ("CH1", "CH2"):
            raise TypeError(f"Analog gain is not available on {self._name}.")

        if value not in GAIN_VALUES:
            raise ValueError(f"Invalid gain. Valid values are {GAIN_VALUES}.")

        self._gain = value
        self._calibrate()

    @property
    def resolution(self) -> int:
        """int: Resolution in bits.

        Setting a new resolution will automatically recalibrate the channel.

        Raises
        ------
        ValueError
            If a resolution other than 10 or 12 is set.
        """
        return int(np.log2((self._resolution + 1)))

    @resolution.setter
    def resolution(self, value: int):
        if value not in (10, 12):
            raise ValueError("Resolution must be 10 or 12 bits.")
        self._resolution = 2 ** value - 1
        self._calibrate()

    def _calibrate(self):
        A = INPUT_RANGES[self._name][0] / self._gain
        B = INPUT_RANGES[self._name][1] / self._gain
        slope = B - A
        intercept = A
        self._scale = np.poly1d([slope / self._resolution, intercept])
        self._unscale = np.poly1d(
            [self._resolution / slope, -self._resolution * intercept / slope]
        )

    def scale(self, raw: Union[int, List[int]]) -> float:
        """Translate raw integer value from device to corresponding voltage.

        Inverse of :meth:`unscale`.

        Parameters
        ----------
        raw : int, List[int]
            An integer, or a list of integers, received from the device.

        Returns
        -------
        float
            Voltage, translated from raw based on channel range, gain, and resolution.
        """
        return self._scale(raw)

    def unscale(self, voltage: float) -> int:
        """Translate a voltage to a raw integer value interpretable by the device.

        Inverse of :meth:`scale`.

        Parameters
        ----------
        voltage : float
            Voltage in volts.

        Returns
        -------
        int
            Corresponding integer value, adjusted for resolution and gain and clipped
            to the channel's range.
        """
        level = self._unscale(voltage)
        level = np.clip(level, 0, self._resolution)
        level = np.round(level)
        return int(level)


class AnalogOutput:
    """Model of the PSLab's analog outputs.

    Parameters
    ----------
    name : str
        Name of the analog output pin represented by this instance.

    Attributes
    ----------
    frequency : float
        Frequency of the waveform on this pin in Hz.
    wavetype : {'sine', 'tria', 'custom'}
        Type of waveform on this pin. 'sine' is a sine wave with amplitude
        3.3 V, 'tria' is a triangle wave with amplitude 3.3 V, 'custom' is any
        other waveform set with :meth:`load_function` or :meth:`load_table`.
    """

    RANGE = (-3.3, 3.3)

    def __init__(self, name):
        self.name = name
        self.frequency = 0
        self.wavetype = "sine"
        self._waveform_table = self.RANGE[1] * np.sin(
            np.arange(
                self.RANGE[0], self.RANGE[1], (self.RANGE[1] - self.RANGE[0]) / 512
            )
        )

    @property
    def waveform_table(self) -> np.ndarray:
        """numpy.ndarray: 512-value waveform table loaded on this output."""
        # A form of amplitude control. Max PWM duty cycle out of 512 clock cycles.
        return self._range_normalize(self._waveform_table, 511)

    @waveform_table.setter
    def waveform_table(self, points: np.ndarray):
        if max(points) - min(points) > self.RANGE[1] - self.RANGE[0]:
            logger.warning(f"Analog output {self.name} saturated.")
        self._waveform_table = np.clip(points, self.RANGE[0], self.RANGE[1])

    @property
    def lowres_waveform_table(self) -> np.ndarray:
        """numpy.ndarray: 32-value waveform table loaded on this output."""
        # Max PWM duty cycle out of 64 clock  cycles.
        return self._range_normalize(self._waveform_table[::16], 63)

    def _range_normalize(self, x: np.ndarray, norm: int = 1) -> np.ndarray:
        """Normalize waveform table to the digital output range."""
        x = (x - self.RANGE[0]) / (self.RANGE[1] - self.RANGE[0]) * norm
        return np.int16(np.round(x)).tolist()



"""TODO"""

import struct


# allows to pack numeric values into byte strings
Byte = struct.Struct("B")  # size 1
ShortInt = struct.Struct("H")  # size 2
Integer = struct.Struct("I")  # size 4

ACKNOWLEDGE = Byte.pack(254)
MAX_SAMPLES = 10000
DATA_SPLITTING = 200
CLOCK_RATE = 64e6

# /*----flash memory----*/
# FLASH = Byte.pack(1)
# READ_FLASH = Byte.pack(1)
# WRITE_FLASH = Byte.pack(2)
# WRITE_BULK_FLASH = Byte.pack(3)
# READ_BULK_FLASH = Byte.pack(4)

# /*-----ADC------*/
ADC = Byte.pack(2)
CAPTURE_ONE = Byte.pack(1)
CAPTURE_TWO = Byte.pack(2)
CAPTURE_DMASPEED = Byte.pack(3)
CAPTURE_FOUR = Byte.pack(4)
CONFIGURE_TRIGGER = Byte.pack(5)
GET_CAPTURE_STATUS = Byte.pack(6)
GET_CAPTURE_CHANNEL = Byte.pack(7)
SET_PGA_GAIN = Byte.pack(8)
GET_VOLTAGE = Byte.pack(9)
GET_VOLTAGE_SUMMED = Byte.pack(10)
# START_ADC_STREAMING = Byte.pack(11)
SELECT_PGA_CHANNEL = Byte.pack(12)
CAPTURE_12BIT = Byte.pack(13)
# CAPTURE_MULTIPLE = Byte.pack(14)
SET_HI_CAPTURE = Byte.pack(15)
SET_LO_CAPTURE = Byte.pack(16)

MULTIPOINT_CAPACITANCE = Byte.pack(20)
SET_CAP = Byte.pack(21)
PULSE_TRAIN = Byte.pack(22)

# /*-----SPI--------*/
SPI_HEADER = Byte.pack(3)
START_SPI = Byte.pack(1)
SEND_SPI8 = Byte.pack(2)
SEND_SPI16 = Byte.pack(3)
STOP_SPI = Byte.pack(4)
SET_SPI_PARAMETERS = Byte.pack(5)
SEND_SPI8_BURST = Byte.pack(6)
SEND_SPI16_BURST = Byte.pack(7)
# /*------I2C-------*/
I2C_HEADER = Byte.pack(4)
I2C_START = Byte.pack(1)
I2C_SEND = Byte.pack(2)
I2C_STOP = Byte.pack(3)
I2C_RESTART = Byte.pack(4)
I2C_READ_END = Byte.pack(5)
I2C_READ_MORE = Byte.pack(6)
I2C_WAIT = Byte.pack(7)
I2C_SEND_BURST = Byte.pack(8)
I2C_CONFIG = Byte.pack(9)
I2C_STATUS = Byte.pack(10)
I2C_READ_BULK = Byte.pack(11)
I2C_WRITE_BULK = Byte.pack(12)
I2C_ENABLE_SMBUS = Byte.pack(13)
I2C_INIT = Byte.pack(14)
I2C_PULLDOWN_SCL = Byte.pack(15)
I2C_DISABLE_SMBUS = Byte.pack(16)

# /*------UART2--------*/
UART_2 = Byte.pack(5)
SEND_BYTE = Byte.pack(1)
SEND_INT = Byte.pack(2)
SEND_ADDRESS = Byte.pack(3)
SET_BAUD = Byte.pack(4)
SET_MODE = Byte.pack(5)
READ_BYTE = Byte.pack(6)
READ_INT = Byte.pack(7)
READ_UART2_STATUS = Byte.pack(8)

# /*-----------DAC--------*/
DAC = Byte.pack(6)
SET_DAC = Byte.pack(1)
SET_CALIBRATED_DAC = Byte.pack(2)

# /*--------WAVEGEN-----*/
WAVEGEN = Byte.pack(7)
SET_WG = Byte.pack(1)
SET_SQR1 = Byte.pack(3)
SET_SQR2 = Byte.pack(4)
SET_SQRS = Byte.pack(5)
TUNE_SINE_OSCILLATOR = Byte.pack(6)
SQR4 = Byte.pack(7)
MAP_REFERENCE = Byte.pack(8)
SET_BOTH_WG = Byte.pack(9)
SET_WAVEFORM_TYPE = Byte.pack(10)
SELECT_FREQ_REGISTER = Byte.pack(11)
DELAY_GENERATOR = Byte.pack(12)
SET_SINE1 = Byte.pack(13)
SET_SINE2 = Byte.pack(14)

LOAD_WAVEFORM1 = Byte.pack(15)
LOAD_WAVEFORM2 = Byte.pack(16)
SQR1_PATTERN = Byte.pack(17)
# /*-----digital outputs----*/
DOUT = Byte.pack(8)
SET_STATE = Byte.pack(1)

# /*-----digital inputs-----*/
DIN = Byte.pack(9)
GET_STATE = Byte.pack(1)
GET_STATES = Byte.pack(2)

LA1 = Byte.pack(0)
LA2 = Byte.pack(1)
LA3 = Byte.pack(2)
LA4 = Byte.pack(3)
LMETER = Byte.pack(4)

# /*------TIMING FUNCTIONS-----*/
TIMING = Byte.pack(10)
GET_TIMING = Byte.pack(1)
GET_PULSE_TIME = Byte.pack(2)
GET_DUTY_CYCLE = Byte.pack(3)
START_ONE_CHAN_LA = Byte.pack(4)
START_TWO_CHAN_LA = Byte.pack(5)
START_FOUR_CHAN_LA = Byte.pack(6)
FETCH_DMA_DATA = Byte.pack(7)
FETCH_INT_DMA_DATA = Byte.pack(8)
FETCH_LONG_DMA_DATA = Byte.pack(9)
GET_LA_PROGRESS = Byte.pack(10)
GET_INITIAL_DIGITAL_STATES = Byte.pack(11)

TIMING_MEASUREMENTS = Byte.pack(12)
INTERVAL_MEASUREMENTS = Byte.pack(13)
CONFIGURE_COMPARATOR = Byte.pack(14)
START_ALTERNATE_ONE_CHAN_LA = Byte.pack(15)
START_THREE_CHAN_LA = Byte.pack(16)
STOP_LA = Byte.pack(17)

# /*--------MISCELLANEOUS------*/
COMMON = Byte.pack(11)

GET_CTMU_VOLTAGE = Byte.pack(1)
GET_CAPACITANCE = Byte.pack(2)
GET_FREQUENCY = Byte.pack(3)
GET_INDUCTANCE = Byte.pack(4)

GET_VERSION = Byte.pack(5)

RETRIEVE_BUFFER = Byte.pack(8)
GET_HIGH_FREQUENCY = Byte.pack(9)
CLEAR_BUFFER = Byte.pack(10)
SET_RGB1 = Byte.pack(11)
READ_PROGRAM_ADDRESS = Byte.pack(12)
WRITE_PROGRAM_ADDRESS = Byte.pack(13)
READ_DATA_ADDRESS = Byte.pack(14)
WRITE_DATA_ADDRESS = Byte.pack(15)

GET_CAP_RANGE = Byte.pack(16)
SET_RGB2 = Byte.pack(17)
READ_LOG = Byte.pack(18)
RESTORE_STANDALONE = Byte.pack(19)
GET_ALTERNATE_HIGH_FREQUENCY = Byte.pack(20)
SET_RGB3 = Byte.pack(22)

START_CTMU = Byte.pack(23)
STOP_CTMU = Byte.pack(24)

START_COUNTING = Byte.pack(25)
FETCH_COUNT = Byte.pack(26)
FILL_BUFFER = Byte.pack(27)

# /*---------- BAUDRATE for main comm channel----*/
SETBAUD = Byte.pack(12)
BAUD9600 = Byte.pack(1)
BAUD14400 = Byte.pack(2)
BAUD19200 = Byte.pack(3)
BAUD28800 = Byte.pack(4)
BAUD38400 = Byte.pack(5)
BAUD57600 = Byte.pack(6)
BAUD115200 = Byte.pack(7)
BAUD230400 = Byte.pack(8)
BAUD1000000 = Byte.pack(9)

# /*-----------NRFL01 radio module----------*/
NRFL01 = Byte.pack(13)
NRF_SETUP = Byte.pack(1)
NRF_RXMODE = Byte.pack(2)
NRF_TXMODE = Byte.pack(3)
NRF_POWER_DOWN = Byte.pack(4)
NRF_RXCHAR = Byte.pack(5)
NRF_TXCHAR = Byte.pack(6)
NRF_HASDATA = Byte.pack(7)
NRF_FLUSH = Byte.pack(8)
NRF_WRITEREG = Byte.pack(9)
NRF_READREG = Byte.pack(10)
NRF_GETSTATUS = Byte.pack(11)
NRF_WRITECOMMAND = Byte.pack(12)
NRF_WRITEPAYLOAD = Byte.pack(13)
NRF_READPAYLOAD = Byte.pack(14)
NRF_WRITEADDRESS = Byte.pack(15)
NRF_TRANSACTION = Byte.pack(16)
NRF_START_TOKEN_MANAGER = Byte.pack(17)
NRF_STOP_TOKEN_MANAGER = Byte.pack(18)
NRF_TOTAL_TOKENS = Byte.pack(19)
NRF_REPORTS = Byte.pack(20)
NRF_WRITE_REPORT = Byte.pack(21)
NRF_DELETE_REPORT_ROW = Byte.pack(22)

NRF_WRITEADDRESSES = Byte.pack(23)

# ---------Non standard IO protocols--------

NONSTANDARD_IO = Byte.pack(14)
# HX711_HEADER = Byte.pack(1)
HCSR04_HEADER = Byte.pack(2)
# AM2302_HEADER = Byte.pack(3)
# TCD1304_HEADER = Byte.pack(4)
# STEPPER_MOTOR = Byte.pack(5)

# --------COMMUNICATION PASSTHROUGHS--------
# Data sent to the device is directly routed to output ports such as (SCL, SDA for UART)

PASSTHROUGHS = Byte.pack(15)
PASS_UART = Byte.pack(1)

# /*--------STOP STREAMING------*/
# STOP_STREAMING = Byte.pack(253)

# /*------INPUT CAPTURE---------*/
# capture modes
EVERY_SIXTEENTH_RISING_EDGE = Byte.pack(0b101)
EVERY_FOURTH_RISING_EDGE = Byte.pack(0b100)
EVERY_RISING_EDGE = Byte.pack(0b011)
EVERY_FALLING_EDGE = Byte.pack(0b010)
EVERY_EDGE = Byte.pack(0b001)
DISABLED = Byte.pack(0b000)

# /*--------Chip selects-----------*/
CSA1 = Byte.pack(1)
CSA2 = Byte.pack(2)
CSA3 = Byte.pack(3)
CSA4 = Byte.pack(4)
CSA5 = Byte.pack(5)
CS1 = Byte.pack(6)
CS2 = Byte.pack(7)

# resolutions
TEN_BIT = Byte.pack(10)
TWELVE_BIT = Byte.pack(12)



import numpy as np


class ComplementaryFilter:
    def __init__(self, ):
        self.pitch = 0
        self.roll = 0
        self.dt = 0.001

    def addData(self, accData, gyrData):
        self.pitch += (gyrData[0]) * self.dt  # Angle around the X-axis
        self.roll -= (gyrData[1]) * self.dt  # Angle around the Y-axis
        forceMagnitudeApprox = abs(accData[0]) + abs(accData[1]) + abs(accData[2]);
        pitchAcc = np.arctan2(accData[1], accData[2]) * 180 / np.pi
        self.pitch = self.pitch * 0.98 + pitchAcc * 0.02
        rollAcc = np.arctan2(accData[0], accData[2]) * 180 / np.pi
        self.roll = self.roll * 0.98 + rollAcc * 0.02

    def getData(self):
        return self.roll, self.pitch



# Registers adapted from sample code for SEMTECH SX1276
from __future__ import print_function

import time


def connect(SPI, frq, **kwargs):
    return SX1276(SPI, frq, **kwargs)


class SX1276():
    name = 'SX1276'
    # registers
    REG_FIFO = 0x00
    REG_OP_MODE = 0x01
    REG_FRF_MSB = 0x06
    REG_FRF_MID = 0x07
    REG_FRF_LSB = 0x08
    REG_PA_CONFIG = 0x09
    REG_LNA = 0x0c
    REG_FIFO_ADDR_PTR = 0x0d
    REG_FIFO_TX_BASE_ADDR = 0x0e
    REG_FIFO_RX_BASE_ADDR = 0x0f
    REG_FIFO_RX_CURRENT_ADDR = 0x10
    REG_IRQ_FLAGS = 0x12
    REG_RX_NB_BYTES = 0x13
    REG_PKT_RSSI_VALUE = 0x1a
    REG_PKT_SNR_VALUE = 0x1b
    REG_MODEM_CONFIG_1 = 0x1d
    REG_MODEM_CONFIG_2 = 0x1e
    REG_PREAMBLE_MSB = 0x20
    REG_PREAMBLE_LSB = 0x21
    REG_PAYLOAD_LENGTH = 0x22
    REG_MODEM_CONFIG_3 = 0x26
    REG_RSSI_WIDEBAND = 0x2c
    REG_DETECTION_OPTIMIZE = 0x31
    REG_DETECTION_THRESHOLD = 0x37
    REG_SYNC_WORD = 0x39
    REG_DIO_MAPPING_1 = 0x40
    REG_VERSION = 0x42
    REG_PA_DAC = 0x4D
    # modes
    MODE_LONG_RANGE_MODE = 0x80
    MODE_SLEEP = 0x00
    MODE_STDBY = 0x01
    MODE_TX = 0x03
    MODE_RX_CONTINUOUS = 0x05
    MODE_RX_SINGLE = 0x06

    # PA config
    PA_BOOST = 0x80

    # IRQ masks
    IRQ_TX_DONE_MASK = 0x08
    IRQ_PAYLOAD_CRC_ERROR_MASK = 0x20
    IRQ_RX_DONE_MASK = 0x40

    MAX_PKT_LENGTH = 255

    PA_OUTPUT_RFO_PIN = 0
    PA_OUTPUT_PA_BOOST_PIN = 1
    _onReceive = 0
    _frequency = 10
    _packetIndex = 0
    packetLength = 0

    def __init__(self, SPI, frq, **kwargs):
        self.SPI = SPI
        self.SPI.set_parameters(2, 6, 1, 0)
        self.name = 'SX1276'
        self.frequency = frq

        self.reset()
        self.version = self.SPIRead(self.REG_VERSION, 1)[0]
        if self.version != 0x12:
            print('version error', self.version)
        self.sleep()
        self.setFrequency(self.frequency)

        # set base address
        self.SPIWrite(self.REG_FIFO_TX_BASE_ADDR, [0])
        self.SPIWrite(self.REG_FIFO_RX_BASE_ADDR, [0])

        # set LNA boost
        self.SPIWrite(self.REG_LNA, [self.SPIRead(self.REG_LNA)[0] | 0x03])

        # set auto ADC
        self.SPIWrite(self.REG_MODEM_CONFIG_3, [0x04])

        # output power 17dbm
        self.setTxPower(kwargs.get('power', 17),
                        self.PA_OUTPUT_PA_BOOST_PIN if kwargs.get('boost', True) else self.PA_OUTPUT_RFO_PIN)
        self.idle()

        # set bandwidth
        self.setSignalBandwidth(kwargs.get('BW', 125e3))
        self.setSpreadingFactor(kwargs.get('SF', 12))
        self.setCodingRate4(kwargs.get('CF', 5))

    def beginPacket(self, implicitHeader=False):
        self.idle()
        if implicitHeader:
            self.implicitHeaderMode()
        else:
            self.explicitHeaderMode()

        # reset FIFO & payload length
        self.SPIWrite(self.REG_FIFO_ADDR_PTR, [0])
        self.SPIWrite(self.REG_PAYLOAD_LENGTH, [0])

    def endPacket(self):
        # put in TX mode
        self.SPIWrite(self.REG_OP_MODE, [self.MODE_LONG_RANGE_MODE | self.MODE_TX])
        while 1:  # Wait for TX done
            if self.SPIRead(self.REG_IRQ_FLAGS, 1)[0] & self.IRQ_TX_DONE_MASK:
                break
            else:
                print('wait...')
                time.sleep(0.1)
        self.SPIWrite(self.REG_IRQ_FLAGS, [self.IRQ_TX_DONE_MASK])

    def parsePacket(self, size=0):
        self.packetLength = 0
        irqFlags = self.SPIRead(self.REG_IRQ_FLAGS, 1)[0]
        if size > 0:
            self.implicitHeaderMode()
            self.SPIWrite(self.REG_PAYLOAD_LENGTH, [size & 0xFF])
        else:
            self.explicitHeaderMode()
        self.SPIWrite(self.REG_IRQ_FLAGS, [irqFlags])
        if (irqFlags & self.IRQ_RX_DONE_MASK) and (irqFlags & self.IRQ_PAYLOAD_CRC_ERROR_MASK) == 0:
            self._packetIndex = 0
            if self._implicitHeaderMode:
                self.packetLength = self.SPIRead(self.REG_PAYLOAD_LENGTH, 1)[0]
            else:
                self.packetLength = self.SPIRead(self.REG_RX_NB_BYTES, 1)[0]
            self.SPIWrite(self.REG_FIFO_ADDR_PTR, self.SPIRead(self.REG_FIFO_RX_CURRENT_ADDR, 1))
            self.idle()
        elif self.SPIRead(self.REG_OP_MODE)[0] != (self.MODE_LONG_RANGE_MODE | self.MODE_RX_SINGLE):
            self.SPIWrite(self.REG_FIFO_ADDR_PTR, [0])
            self.SPIWrite(self.REG_OP_MODE, [self.MODE_LONG_RANGE_MODE | self.MODE_RX_SINGLE])
        return self.packetLength

    def packetRssi(self):
        return self.SPIRead(self.REG_PKT_RSSI_VALUE)[0] - (164 if self._frequency < 868e6 else 157)

    def packetSnr(self):
        return self.SPIRead(self.REG_PKT_SNR_VALUE)[0] * 0.25

    def write(self, byteArray):
        size = len(byteArray)
        currentLength = self.SPIRead(self.REG_PAYLOAD_LENGTH)[0]
        if (currentLength + size) > self.MAX_PKT_LENGTH:
            size = self.MAX_PKT_LENGTH - currentLength
        self.SPIWrite(self.REG_FIFO, byteArray[:size])
        self.SPIWrite(self.REG_PAYLOAD_LENGTH, [currentLength + size])
        return size

    def available(self):
        return self.SPIRead(self.REG_RX_NB_BYTES)[0] - self._packetIndex

    def checkRx(self):
        irqFlags = self.SPIRead(self.REG_IRQ_FLAGS, 1)[0]
        if (irqFlags & self.IRQ_RX_DONE_MASK) and (irqFlags & self.IRQ_PAYLOAD_CRC_ERROR_MASK) == 0:
            return 1
        return 0;

    def read(self):
        if not self.available(): return -1
        self._packetIndex += 1
        return self.SPIRead(self.REG_FIFO)[0]

    def readAll(self):
        p = []
        while self.available():
            p.append(self.read())
        return p

    def peek(self):
        if not self.available(): return -1
        self.currentAddress = self.SPIRead(self.REG_FIFO_ADDR_PTR)
        val = self.SPIRead(self.REG_FIFO)[0]
        self.SPIWrite(self.REG_FIFO_ADDR_PTR, self.currentAddress)
        return val

    def flush(self):
        pass

    def receive(self, size):
        if size > 0:
            self.implicitHeaderMode()
            self.SPIWrite(self.REG_PAYLOAD_LENGTH, [size & 0xFF])
        else:
            self.explicitHeaderMode()

        self.SPIWrite(self.REG_OP_MODE, [self.MODE_LONG_RANGE_MODE | self.MODE_RX_SINGLE])

    def reset(self):
        pass

    def idle(self):
        self.SPIWrite(self.REG_OP_MODE, [self.MODE_LONG_RANGE_MODE | self.MODE_STDBY])

    def sleep(self):
        self.SPIWrite(self.REG_OP_MODE, [self.MODE_LONG_RANGE_MODE | self.MODE_SLEEP])

    def setTxPower(self, level, pin):
        if pin == self.PA_OUTPUT_RFO_PIN:
            if level < 0:
                level = 0
            elif level > 14:
                level = 14
            self.SPIWrite(self.REG_PA_CONFIG, [0x70 | level])
        else:
            if level < 2:
                level = 2
            elif level > 17:
                level = 17
            if level == 17:
                print('max power output')
                self.SPIWrite(self.REG_PA_DAC, [0x87])
            else:
                self.SPIWrite(self.REG_PA_DAC, [0x84])
            self.SPIWrite(self.REG_PA_CONFIG, [self.PA_BOOST | 0x70 | (level - 2)])

        print('power', hex(self.SPIRead(self.REG_PA_CONFIG)[0]))

    def setFrequency(self, frq):
        self._frequency = frq
        frf = (int(frq) << 19) / 32000000
        print('frf', frf)
        print('freq', (frf >> 16) & 0xFF, (frf >> 8) & 0xFF, (frf) & 0xFF)
        self.SPIWrite(self.REG_FRF_MSB, [(frf >> 16) & 0xFF])
        self.SPIWrite(self.REG_FRF_MID, [(frf >> 8) & 0xFF])
        self.SPIWrite(self.REG_FRF_LSB, [frf & 0xFF])

    def setSpreadingFactor(self, sf):
        if sf < 6:
            sf = 6
        elif sf > 12:
            sf = 12

        if sf == 6:
            self.SPIWrite(self.REG_DETECTION_OPTIMIZE, [0xc5])
            self.SPIWrite(self.REG_DETECTION_THRESHOLD, [0x0c])
        else:
            self.SPIWrite(self.REG_DETECTION_OPTIMIZE, [0xc3])
            self.SPIWrite(self.REG_DETECTION_THRESHOLD, [0x0a])
        self.SPIWrite(self.REG_MODEM_CONFIG_2, [(self.SPIRead(self.REG_MODEM_CONFIG_2)[0] & 0x0F) | ((sf << 4) & 0xF0)])

    def setSignalBandwidth(self, sbw):
        bw = 9
        num = 0
        for a in [7.8e3, 10.4e3, 15.6e3, 20.8e3, 31.25e3, 41.7e3, 62.5e3, 125e3, 250e3]:
            if sbw <= a:
                bw = num
                break
            num += 1
        print('bandwidth: ', bw)
        self.SPIWrite(self.REG_MODEM_CONFIG_1, [(self.SPIRead(self.REG_MODEM_CONFIG_1)[0] & 0x0F) | (bw << 4)])

    def setCodingRate4(self, denominator):
        if denominator < 5:
            denominator = 5
        elif denominator > 8:
            denominator = 8
        self.SPIWrite(self.REG_MODEM_CONFIG_1,
                      [(self.SPIRead(self.REG_MODEM_CONFIG_1)[0] & 0xF1) | ((denominator - 4) << 4)])

    def setPreambleLength(self, length):
        self.SPIWrite(self.REG_PREAMBLE_MSB, [(length >> 8) & 0xFF])
        self.SPIWrite(self.REG_PREAMBLE_LSB, [length & 0xFF])

    def setSyncWord(self, sw):
        self.SPIWrite(self.REG_SYNC_WORD, [sw])

    def crc(self):
        self.SPIWrite(self.REG_MODEM_CONFIG_2, [self.SPIRead(self.REG_MODEM_CONFIG_2)[0] | 0x04])

    def noCrc(self):
        self.SPIWrite(self.REG_MODEM_CONFIG_2, [self.SPIRead(self.REG_MODEM_CONFIG_2)[0] & 0xFB])

    def random(self):
        return self.SPIRead(self.REG_RSSI_WIDEBAND)[0]

    def explicitHeaderMode(self):
        self._implicitHeaderMode = 0
        self.SPIWrite(self.REG_MODEM_CONFIG_1, [self.SPIRead(self.REG_MODEM_CONFIG_1)[0] & 0xFE])

    def implicitHeaderMode(self):
        self._implicitHeaderMode = 1
        self.SPIWrite(self.REG_MODEM_CONFIG_1, [self.SPIRead(self.REG_MODEM_CONFIG_1)[0] | 0x01])

    def handleDio0Rise(self):
        irqFlags = self.SPIRead(self.REG_IRQ_FLAGS, 1)[0]
        self.SPIWrite(self.REG_IRQ_FLAGS, [irqFlags])

        if (irqFlags & self.IRQ_PAYLOAD_CRC_ERROR_MASK) == 0:
            self._packetIndex = 0
            if self._implicitHeaderMode:
                self.packetLength = self.SPIRead(self.REG_PAYLOAD_LENGTH, 1)[0]
            else:
                self.packetLength = self.SPIRead(self.REG_RX_NB_BYTES, 1)[0]

            self.SPIWrite(self.REG_FIFO_ADDR_PTR, self.SPIRead(self.REG_FIFO_RX_CURRENT_ADDR, 1))
            if self._onReceive:
                print(self.packetLength)
        # self._onReceive(self.packetLength)

        self.SPIWrite(self.REG_FIFO_ADDR_PTR, [0])

    def SPIWrite(self, adr, byteArray):
        return self.SPI.xfer('CS1', [0x80 | adr] + byteArray)[1:]

    def SPIRead(self, adr, total_bytes=1):
        return self.SPI.xfer('CS1', [adr] + [0] * total_bytes)[1:]

    def getRaw(self):
        val = self.SPIRead(0x02, 1)
        return val


if __name__ == "__main__":
    RX = 0;
    TX = 1
    mode = RX
    from PSL import sciencelab

    I = sciencelab.connect()
    lora = SX1276(I.SPI, 434e6, boost=True, power=17, BW=125e3, SF=12, CR=5)  # settings for maximum range
    lora.crc()
    cntr = 0
    while 1:
        time.sleep(0.01)
        if mode == TX:
            lora.beginPacket()
            lora.write([cntr])
            # lora.write([ord(a) for a in ":"]+[cntr])
            print(time.ctime(), [ord(a) for a in ":"] + [cntr], hex(lora.SPIRead(lora.REG_OP_MODE)[0]))
            lora.endPacket()
            cntr += 1
            if cntr == 255: cntr = 0
        elif mode == RX:
            packet_size = lora.parsePacket()
            if packet_size:
                print('data', lora.readAll())
                print('Rssi', lora.packetRssi(), lora.packetSnr())



"""Contains modules for controlling the PSLab's built-in instruments.

The built-in instruments are ``LogicAnalyzer``, ``Multimeter``,
``Oscilloscope``, ``PowerSupply``, ``PWMGenerator``, and ``WaveformGenerator``.
"""



import sys


class AD9833:
    if sys.version.major == 3:
        DDS_MAX_FREQ = 0xFFFFFFF - 1  # 24 bit resolution
    else:
        DDS_MAX_FREQ = eval("0xFFFFFFFL-1")  # 24 bit resolution
    # control bytes
    DDS_B28 = 13
    DDS_HLB = 12
    DDS_FSELECT = 11
    DDS_PSELECT = 10
    DDS_RESET = 8
    DDS_SLEEP1 = 7
    DDS_SLEEP12 = 6
    DDS_OPBITEN = 5
    DDS_DIV2 = 3
    DDS_MODE = 1

    DDS_FSYNC = 9

    DDS_SINE = (0)
    DDS_TRIANGLE = (1 << DDS_MODE)
    DDS_SQUARE = (1 << DDS_OPBITEN)
    DDS_RESERVED = (1 << DDS_OPBITEN) | (1 << DDS_MODE)
    clockScaler = 4  # 8MHz

    def __init__(self, I=None):
        self.CS = 9
        if I:
            self.I = I
        else:
            from PSL import sciencelab
            self.I = sciencelab.connect()
        self.I.SPI.set_parameters(2, 2, 1, 1, 0)
        self.I.map_reference_clock(self.clockScaler, 'WAVEGEN')
        print('clock set to ', self.I.DDS_CLOCK)

        self.waveform_mode = self.DDS_TRIANGLE;
        self.write(1 << self.DDS_RESET)
        self.write((1 << self.DDS_B28) | self.waveform_mode)  # finished loading data
        self.active_channel = 0
        self.frequency = 1000

    def write(self, con):
        self.I.SPI.start(self.CS)
        self.I.SPI.send16(con)
        self.I.SPI.stop(self.CS)

    def set_frequency(self, freq, register=0, **args):
        self.active_channel = register
        self.frequency = freq

        freq_setting = int(round(freq * self.DDS_MAX_FREQ / self.I.DDS_CLOCK))
        modebits = (1 << self.DDS_B28) | self.waveform_mode
        if register:
            modebits |= (1 << self.DDS_FSELECT)
            regsel = 0x8000
        else:
            regsel = 0x4000

        self.write((1 << self.DDS_RESET) | modebits)  # Ready to load DATA
        self.write((regsel | (freq_setting & 0x3FFF)) & 0xFFFF)  # LSB
        self.write((regsel | ((freq_setting >> 14) & 0x3FFF)) & 0xFFFF)  # MSB
        phase = args.get('phase', 0)
        self.write(0xc000 | phase)  # Phase
        self.write(modebits)  # finished loading data

    def set_voltage(self, v):
        self.waveform_mode = self.DDS_TRIANGLE
        self.set_frequency(0, 0, phase=v)  # 0xfff*v/.6)

    def select_frequency_register(self, register):
        self.active_channel = register
        modebits = self.waveform_mode
        if register:    modebits |= (1 << self.DDS_FSELECT)
        self.write(modebits)

    def set_waveform_mode(self, mode):
        self.waveform_mode = mode
        modebits = mode
        if self.active_channel:    modebits |= (1 << self.DDS_FSELECT)
        self.write(modebits)


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    A = AD9833(I=I)
    A.set_waveform_mode(A.DDS_SINE)
    A.set_frequency(3600, 0)
    A.set_frequency(3600, 1)



def connect(route, **args):
    return HMC5883L(route, **args)


class HMC5883L():
    CONFA = 0x00
    CONFB = 0x01
    MODE = 0x02
    STATUS = 0x09

    # --------CONFA register bits. 0x00-----------
    samplesToAverage = 0
    samplesToAverage_choices = [1, 2, 4, 8]

    dataOutputRate = 6
    dataOutputRate_choices = [0.75, 1.5, 3, 7.5, 15, 30, 75]

    measurementConf = 0

    # --------CONFB register bits. 0x01-----------
    gainValue = 7  # least sensitive
    gain_choices = [8, 7, 6, 5, 4, 3, 2, 1]
    scaling = [1370., 1090., 820., 660., 440., 390., 330., 230.]

    # --------------Parameters--------------------
    # This must be defined in order to let GUIs automatically create menus
    # for changing various options of this sensor
    # It's a dictionary of the string representations of functions matched with an array
    # of options that each one can accept
    params = {'init': None,
              'setSamplesToAverage': samplesToAverage_choices,
              'setDataOutputRate': dataOutputRate_choices,
              'setGain': gain_choices,
              }
    ADDRESS = 0x1E
    name = 'Magnetometer'
    NUMPLOTS = 3
    PLOTNAMES = ['Bx', 'By', 'Bz']

    def __init__(self, I2C, **args):
        self.I2C = I2C
        self.ADDRESS = args.get('address', self.ADDRESS)
        self.name = 'Magnetometer'
        '''
        try:
            print 'switching baud to 400k'
            self.I2C.configI2C(400e3)
        except:
            print 'FAILED TO CHANGE BAUD RATE'
        '''
        self.init()

    def init(self):
        self.__writeCONFA__()
        self.__writeCONFB__()
        self.I2C.writeBulk(self.ADDRESS, [self.MODE, 0])  # enable continuous measurement mode

    def __writeCONFB__(self):
        self.I2C.writeBulk(self.ADDRESS, [self.CONFB, self.gainValue << 5])  # set gain

    def __writeCONFA__(self):
        self.I2C.writeBulk(self.ADDRESS, [self.CONFA, (self.dataOutputRate << 2) | (self.samplesToAverage << 5) | (
            self.measurementConf)])

    def setSamplesToAverage(self, num):
        self.samplesToAverage = self.samplesToAverage_choices.index(num)
        self.__writeCONFA__()

    def setDataOutputRate(self, rate):
        self.dataOutputRate = self.dataOutputRate_choices.index(rate)
        self.__writeCONFA__()

    def setGain(self, gain):
        self.gainValue = self.gain_choices.index(gain)
        self.__writeCONFB__()

    def getVals(self, addr, numbytes):
        vals = self.I2C.readBulk(self.ADDRESS, addr, numbytes)
        return vals

    def getRaw(self):
        vals = self.getVals(0x03, 6)
        if vals:
            if len(vals) == 6:
                return [int16(vals[a * 2] << 8 | vals[a * 2 + 1]) / self.scaling[self.gainValue] for a in range(3)]
            else:
                return False
        else:
            return False


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    I.set_sine1(.5)
    A = connect(I.I2C)
    A.setGain(2)
    t, x, y, z = I.I2C.capture(A.ADDRESS, 0x03, 6, 400, 10000, 'int')
    # print (t,x,y,z)
    from pylab import *

    plot(t, x)
    plot(t, y)
    plot(t, z)
    show()



from __future__ import print_function


def connect(route, **args):
    return BRIDGE(route, **args)


class BRIDGE():
    POWER_ON = 0x01
    RESET = 0x07
    RES_1000mLx = 0x10
    RES_500mLx = 0x11
    RES_4000mLx = 0x13

    gain_choices = [RES_500mLx, RES_1000mLx, RES_4000mLx]
    gain_literal_choices = ['500mLx', '1000mLx', '4000mLx']
    gain = 0
    scaling = [2, 1, .25]

    # --------------Parameters--------------------
    # This must be defined in order to let GUIs automatically create menus
    # for changing various options of this sensor
    # It's a dictionary of the string representations of functions matched with an array
    # of options that each one can accept
    params = {'init': None,
              'setRange': gain_literal_choices,
              }

    NUMPLOTS = 1
    PLOTNAMES = ['Lux']
    ADDRESS = 0x23
    name = 'Luminosity'

    def __init__(self, I2C, **args):
        self.I2C = I2C
        self.ADDRESS = args.get('address', 0x23)
        self.init()

    def init(self):
        self.I2C.writeBulk(self.ADDRESS, [self.RES_500mLx])

    def setRange(self, g):
        self.gain = self.gain_literal_choices.index(g)
        self.I2C.writeBulk(self.ADDRESS, [self.gain_choices[self.gain]])

    def getVals(self, numbytes):
        vals = self.I2C.simpleRead(self.ADDRESS, numbytes)
        return vals

    def getRaw(self):
        vals = self.getVals(2)
        if vals:
            if len(vals) == 2:
                return [(vals[0] << 8 | vals[1]) / 1.2]  # /self.scaling[self.gain]
            else:
                return False
        else:
            return False


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    A = connect(I.I2C)
    print(A.getRaw())



from __future__ import print_function
import time


def connect(route, **args):
    '''
    route can either be I.I2C , or a radioLink instance
    '''
    return SHT21(route, **args)


def rawToTemp(vals):
    if vals:
        if len(vals):
            v = (vals[0] << 8) | (vals[1] & 0xFC)  # make integer & remove status bits
            v *= 175.72
            v /= (1 << 16)
            v -= 46.85
            return [v]
    return False


def rawToRH(vals):
    if vals:
        if len(vals):
            v = (vals[0] << 8) | (vals[1] & 0xFC)  # make integer & remove status bits
            v *= 125.
            v /= (1 << 16)
            v -= 6
            return [v]
    return False


class SHT21():
    RESET = 0xFE
    TEMP_ADDRESS = 0xF3
    HUMIDITY_ADDRESS = 0xF5
    selected = 0xF3
    NUMPLOTS = 1
    PLOTNAMES = ['Data']
    ADDRESS = 0x40
    name = 'Humidity/Temperature'

    def __init__(self, I2C, **args):
        self.I2C = I2C
        self.ADDRESS = args.get('address', self.ADDRESS)
        self.name = 'Humidity/Temperature'
        '''
        try:
            print ('switching baud to 400k')
            self.I2C.configI2C(400e3)
        except:
            print ('FAILED TO CHANGE BAUD RATE')
        '''
        self.params = {'selectParameter': ['temperature', 'humidity'], 'init': None}
        self.init()

    def init(self):
        self.I2C.writeBulk(self.ADDRESS, [self.RESET])  # soft reset
        time.sleep(0.1)

    @staticmethod
    def _calculate_checksum(data, number_of_bytes):
        """5.7 CRC Checksum using the polynomial given in the datasheet
        Credits: https://github.com/jaques/sht21_python/blob/master/sht21.py
        """
        # CRC
        POLYNOMIAL = 0x131  # //P(x)=x^8+x^5+x^4+1 = 100110001
        crc = 0
        # calculates 8-Bit checksum with given polynomial
        for byteCtr in range(number_of_bytes):
            crc ^= (data[byteCtr])
            for _ in range(8, 0, -1):
                if crc & 0x80:
                    crc = (crc << 1) ^ POLYNOMIAL
                else:
                    crc = (crc << 1)
        return crc

    def selectParameter(self, param):
        if param == 'temperature':
            self.selected = self.TEMP_ADDRESS
        elif param == 'humidity':
            self.selected = self.HUMIDITY_ADDRESS

    def getRaw(self):
        self.I2C.writeBulk(self.ADDRESS, [self.selected])
        if self.selected == self.TEMP_ADDRESS:
            time.sleep(0.1)
        elif self.selected == self.HUMIDITY_ADDRESS:
            time.sleep(0.05)

        vals = self.I2C.simpleRead(self.ADDRESS, 3)
        if vals:
            if self._calculate_checksum(vals, 2) != vals[2]:
                print(vals)
                return False
        if self.selected == self.TEMP_ADDRESS:
            return rawToTemp(vals)
        elif self.selected == self.HUMIDITY_ADDRESS:
            return rawToRH(vals)



"""Convenience module that creates instances of every instrument for you.

Every PSLab instrument can be imported and instantiated individually. However,
if you need to use several at once the ScienceLab class provides a convenient
collection.
"""
from typing import Iterable, List

import pslab.protocol as CP
from pslab.bus.i2c import I2CMaster
from pslab.instrument.logic_analyzer import LogicAnalyzer
from pslab.instrument.multimeter import Multimeter
from pslab.instrument.oscilloscope import Oscilloscope
from pslab.instrument.power_supply import PowerSupply
from pslab.instrument.waveform_generator import PWMGenerator, WaveformGenerator
from pslab.peripherals import NRF24L01
from pslab.serial_handler import SerialHandler


class ScienceLab(SerialHandler):
    """Aggregate interface for the PSLab's instruments.

    Attributes
    ----------
    logic_analyzer : pslab.LogicAnalyzer
    oscilloscope : pslab.Oscilloscope
    waveform_generator : pslab.WaveformGenerator
    pwm_generator : pslab.PWMGenerator
    multimeter : pslab.Multimeter
    power_supply : pslab.PowerSupply
    i2c : pslab.I2CMaster
    nrf : pslab.peripherals.NRF24L01
    """

    def __init__(
        self,
        port: str = None,
        baudrate: int = 1000000,
        timeout: float = 1.0,
    ):
        super().__init__(port, baudrate, timeout)
        self.logic_analyzer = LogicAnalyzer(device=self)
        self.oscilloscope = Oscilloscope(device=self)
        self.waveform_generator = WaveformGenerator(device=self)
        self.pwm_generator = PWMGenerator(device=self)
        self.multimeter = Multimeter(device=self)
        self.power_supply = PowerSupply(device=self)
        self.i2c = I2CMaster(device=self)
        self.nrf = NRF24L01(device=self)

        if "V6" in self.version:  # Set the built-in WS2812B to green :)
            self.rgb_led([0, 20, 0])

    @property
    def temperature(self):
        """float: Temperature of the MCU in degrees Celsius."""
        # TODO: Get rid of magic numbers.
        cs = 3
        V = self._get_ctmu_voltage(0b11110, cs, 0)

        if cs == 1:
            return (646 - V * 1000) / 1.92  # current source = 1
        elif cs == 2:
            return (701.5 - V * 1000) / 1.74  # current source = 2
        elif cs == 3:
            return (760 - V * 1000) / 1.56  # current source = 3

    def _get_ctmu_voltage(self, channel: int, current_range: int, tgen: bool = True):
        """Control the Charge Time Measurement Unit (CTMU).

        ctmu_voltage(5, 2) will activate a constant current source of 5.5 µA on
        CAP and then measure the voltage at the output.

        If a diode is used to connect CAP to ground, the forward voltage drop
        of the diode will be returned, e.g. 0.6 V for a 4148 diode.

        If a resistor is connected, Ohm's law will be followed within
        reasonable limits.

        Parameters
        ----------
        channel : int
            Pin number on which to generate a current and measure output
            voltage. Refer to the PIC24EP64GP204 datasheet for channel
            numbering.
        current_range : {0, 1, 2, 3}
            0 -> 550 µA
            1 -> 550 nA
            2 -> 5.5 µA
            3 -> 55 µA
        tgen : bool, optional
            Use Time Delay mode instead of Measurement mode. The default value
            is True.

        Returns
        -------
        voltage : float
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.GET_CTMU_VOLTAGE)
        self.send_byte((channel) | (current_range << 5) | (tgen << 7))
        raw_voltage = self.get_int() / 16  # 16*voltage across the current source
        self.get_ack()
        vmax = 3.3
        resolution = 12
        voltage = vmax * raw_voltage / (2 ** resolution - 1)
        return voltage

    def _start_ctmu(self, current_range: int, trim: int, tgen: int = 1):
        self.send_byte(CP.COMMON)
        self.send_byte(CP.START_CTMU)
        self.send_byte((current_range) | (tgen << 7))
        self.send_byte(trim)
        self.get_ack()

    def _stop_ctmu(self):
        self.send_byte(CP.COMMON)
        self.send_byte(CP.STOP_CTMU)
        self.get_ack()

    def reset(self):
        """Reset the device.

        Standalone mode will be enabled if an OLED is connected to the I2C
        port.
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.RESTORE_STANDALONE)

    def rgb_led(self, colors: List, output: str = "RGB", order: str = "GRB"):
        """Set shade of a WS2812B RGB LED.

        Parameters
        ----------
        colors : list
            List of three values between 0-255, where each value is the
            intensity of red, green, and blue, respectively. When daisy
            chaining several LEDs, colors should be a list of three-value
            lists.
        output : {"RGB", "PGC", "SQ1"}, optional
            Pin on which to output the pulse train setting the LED color. The
            default value, "RGB", sets the color of the built-in WS2812B
            (PSLab v6 only).
        order : str, optional
            Color order of the connected LED as a three-letter string. The
            built-in LED has order "GRB", which is the default.

        Examples
        --------
        Set the built-in WS2812B to yellow.

        >>> import pslab
        >>> psl = pslab.ScienceLab()
        >>> psl.rgb_led([10, 10, 0])

        Set a chain of three RGB WS2812B connected to SQ1 to red, cyan, and
        magenta.

        >>> psl.rgb_led([[10,0,0],[0,10,10],[10,0,10]], output="SQ1", order="RGB")
        """
        if output == "RGB":
            pin = CP.SET_RGB1
        elif output == "PGC":
            pin = CP.SET_RGB2
        elif output == "SQ1":
            pin = CP.SET_RGB3
        else:
            raise ValueError(
                f"Invalid output: {output}. output must be 'RGB', 'PCG', or 'SQ1'."
            )

        if not isinstance(colors[0], Iterable):
            colors = [colors]

        if not all([len(color) == 3 for color in colors]):
            raise ValueError("Invalid color; each color list must have three values.")

        order = order.upper()

        if not sorted(order) == ["B", "G", "R"]:
            raise ValueError(
                f"Invalid order: {order}. order must contain 'R', 'G', and 'B'."
            )

        self.send_byte(CP.COMMON)
        self.send_byte(pin)
        self.send_byte(len(colors) * 3)

        for color in colors:
            self.send_byte(color[order.index("R")])
            self.send_byte(color[order.index("G")])
            self.send_byte(color[order.index("B")])

        self.get_ack()

    def _read_program_address(self, address: int):
        """Return the value stored at the specified address in program memory.

        Parameters
        ----------
        address : int
            Address to read from. Refer to PIC24EP64GP204 programming manual.

        Returns
        -------
        data : int
            16-bit wide value read from program memory.
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.READ_PROGRAM_ADDRESS)
        self.send_int(address & 0xFFFF)
        self.send_int((address >> 16) & 0xFFFF)
        data = self.get_int()
        self.get_ack()
        return data

    def _device_id(self):
        a = self._read_program_address(0x800FF8)
        b = self._read_program_address(0x800FFA)
        c = self._read_program_address(0x800FFC)
        d = self._read_program_address(0x800FFE)
        val = d | (c << 16) | (b << 32) | (a << 48)
        return val

    def _read_data_address(self, address: int):
        """Return the value stored at the specified address in RAM.

        Parameters
        ----------
        address : int
            Address to read from. Refer to PIC24EP64GP204 programming manual.

        Returns
        -------
        data : int
            16-bit wide value read from RAM.
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.READ_DATA_ADDRESS)
        self.send_int(address & 0xFFFF)
        data = self.get_int()
        self.get_ack()
        return data

    def _write_data_address(self, address: int, value: int):
        """Write a value to the specified address in RAM.

        Parameters
        ----------
        address : int
            Address to write to.  Refer to PIC24EP64GP204 programming manual.
        value : int
            Value to write to RAM.
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.WRITE_DATA_ADDRESS)
        self.send_int(address & 0xFFFF)
        self.send_int(value)
        self.get_ack()

    def enable_uart_passthrough(self, baudrate: int, persist=False):
        """Relay all data received by the device to TXD/RXD.

        If a period > 0.5 seconds elapses between two transmit/receive events,
        the device resets and resumes normal mode. This timeout feature has
        been implemented in lieu of a hard reset option.

        Can be used to load programs into secondary microcontrollers with
        bootloaders such ATMEGA or ESP8266

        Parameters
        ----------
        baudrate : int
            Baudrate of the UART bus.
        persist : bool, optional
            If set to True, the device will stay in passthrough mode until the
            next power cycle. Otherwise(default scenario), the device will
            return to normal operation if no data is sent/received for a period
            greater than one second at a time.
        """
        self.send_byte(CP.PASSTHROUGHS)
        self.send_byte(CP.PASS_UART)
        self.send_byte(1 if persist else 0)
        self.send_int(int(round(((64e6 / baudrate) / 4) - 1)))

    def set_uart_baud(self, baudrate: int):
        """Set the baudrate of the UART bus.

        Parameters
        ----------
        baudrate : int
            Baudrate to set on the UART bus.
        """
        self.send_byte(CP.UART_2)
        self.send_byte(CP.SET_BAUD)
        self.send_int(int(round(((64e6 / baudrate) / 4) - 1)))
        self.get_ack()

    def write_uart(self, byte: int):
        """Write a single byte to the UART bus.

        Parameters
        ----------
        byte : int
            Byte value to write to the UART bus.
        """
        self.send_byte(CP.UART_2)
        self.send_byte(CP.SEND_BYTE)
        self.send_byte(byte)
        self.get_ack()

    def read_uart(self):
        """Read a single byte from the UART bus.

        Returns
        -------
        byte : int
            Byte value read from the UART bus."""
        self.send_byte(CP.UART_2)
        self.send_byte(CP.READ_BYTE)
        return self.get_byte()

    def read_uart_status(self):
        """Return available bytes in UART buffer.

        Returns
        -------
        status : int
        """
        self.send_byte(CP.UART_2)
        self.send_byte(CP.READ_UART2_STATUS)
        return self.get_byte()

    def read_log(self):
        """Read hardware debug log.

        Returns
        -------
        log : bytes
            Bytes read from the hardware debug log.
        """
        self.send_byte(CP.COMMON)
        self.send_byte(CP.READ_LOG)
        log = self.interface.readline().strip()
        self.get_ack()
        return log



# -*- coding: utf-8; mode: python; indent-tabs-mode: t; tab-width:4 -*-
from Kalman import KalmanFilter


def connect(route, **args):
    return MPU925x(route, **args)


class MPU925x():
    '''
    Mandatory members:
    GetRaw : Function called by Graphical apps. Must return values stored in a list
    NUMPLOTS : length of list returned by GetRaw. Even single datapoints need to be stored in a list before returning
    PLOTNAMES : a list of strings describing each element in the list returned by GetRaw. len(PLOTNAMES) = NUMPLOTS
    name : the name of the sensor shown to the user
    params:
        A dictionary of function calls(single arguments only) paired with list of valid argument values. (Primitive. I know.)
        These calls can be used for one time configuration settings

    '''
    INT_PIN_CFG = 0x37
    GYRO_CONFIG = 0x1B
    ACCEL_CONFIG = 0x1C
    GYRO_SCALING = [131, 65.5, 32.8, 16.4]
    ACCEL_SCALING = [16384, 8192, 4096, 2048]
    AR = 3
    GR = 3
    NUMPLOTS = 7
    PLOTNAMES = ['Ax', 'Ay', 'Az', 'Temp', 'Gx', 'Gy', 'Gz']
    ADDRESS = 0x68
    AK8963_ADDRESS = 0x0C
    AK8963_CNTL = 0x0A
    name = 'Accel/gyro'

    def __init__(self, I2C, **args):
        self.I2C = I2C
        self.ADDRESS = args.get('address', self.ADDRESS)
        self.name = 'Accel/gyro'
        self.params = {'powerUp': None, 'setGyroRange': [250, 500, 1000, 2000], 'setAccelRange': [2, 4, 8, 16],
                       'KalmanFilter': [.01, .1, 1, 10, 100, 1000, 10000, 'OFF']}
        self.setGyroRange(2000)
        self.setAccelRange(16)
        self.powerUp()
        self.K = None

    def KalmanFilter(self, opt):
        if opt == 'OFF':
            self.K = None
            return
        noise = [[]] * self.NUMPLOTS
        for a in range(500):
            vals = self.getRaw()
            for b in range(self.NUMPLOTS): noise[b].append(vals[b])

        self.K = [None] * 7
        for a in range(self.NUMPLOTS):
            sd = std(noise[a])
            self.K[a] = KalmanFilter(1. / opt, sd ** 2)

    def getVals(self, addr, numbytes):
        return self.I2C.readBulk(self.ADDRESS, addr, numbytes)

    def powerUp(self):
        self.I2C.writeBulk(self.ADDRESS, [0x6B, 0])

    def setGyroRange(self, rs):
        self.GR = self.params['setGyroRange'].index(rs)
        self.I2C.writeBulk(self.ADDRESS, [self.GYRO_CONFIG, self.GR << 3])

    def setAccelRange(self, rs):
        self.AR = self.params['setAccelRange'].index(rs)
        self.I2C.writeBulk(self.ADDRESS, [self.ACCEL_CONFIG, self.AR << 3])

    def getRaw(self):
        '''
        This method must be defined if you want GUIs to use this class to generate plots on the fly.
        It must return a set of different values read from the sensor. such as X,Y,Z acceleration.
        The length of this list must not change, and must be defined in the variable NUMPLOTS.

        GUIs will generate as many plots, and the data returned from this method will be appended appropriately
        '''
        vals = self.getVals(0x3B, 14)
        if vals:
            if len(vals) == 14:
                raw = [0] * 7
                for a in range(3): raw[a] = 1. * int16(vals[a * 2] << 8 | vals[a * 2 + 1]) / self.ACCEL_SCALING[self.AR]
                for a in range(4, 7): raw[a] = 1. * int16(vals[a * 2] << 8 | vals[a * 2 + 1]) / self.GYRO_SCALING[
                    self.GR]
                raw[3] = int16(vals[6] << 8 | vals[7]) / 340. + 36.53
                if not self.K:
                    return raw
                else:
                    for b in range(self.NUMPLOTS):
                        self.K[b].input_latest_noisy_measurement(raw[b])
                        raw[b] = self.K[b].get_latest_estimated_measurement()
                    return raw

            else:
                return False
        else:
            return False

    def getAccel(self):
        '''
        Return a list of 3 values for acceleration vector

        '''
        vals = self.getVals(0x3B, 6)
        ax = int16(vals[0] << 8 | vals[1])
        ay = int16(vals[2] << 8 | vals[3])
        az = int16(vals[4] << 8 | vals[5])
        return [ax / 65535., ay / 65535., az / 65535.]

    def getTemp(self):
        '''
        Return temperature
        '''
        vals = self.getVals(0x41, 6)
        t = int16(vals[0] << 8 | vals[1])
        return t / 65535.

    def getGyro(self):
        '''
        Return a list of 3 values for angular velocity vector

        '''
        vals = self.getVals(0x43, 6)
        ax = int16(vals[0] << 8 | vals[1])
        ay = int16(vals[2] << 8 | vals[3])
        az = int16(vals[4] << 8 | vals[5])
        return [ax / 65535., ay / 65535., az / 65535.]

    def getMag(self):
        '''
        Return a list of 3 values for magnetic field vector

        '''
        vals = self.I2C.readBulk(self.AK8963_ADDRESS, 0x03,
                                 7)  # 6+1 . 1(ST2) should not have bit 4 (0x8) true. It's ideally 16 . overflow bit
        ax = int16(vals[0] << 8 | vals[1])
        ay = int16(vals[2] << 8 | vals[3])
        az = int16(vals[4] << 8 | vals[5])
        if not vals[6] & 0x08:
            return [ax / 65535., ay / 65535., az / 65535.]
        else:
            return None

    def WhoAmI(self):
        '''
        Returns the ID.
        It is 71 for MPU9250.
        '''
        v = self.I2C.readBulk(self.ADDRESS, 0x75, 1)[0]
        if v not in [0x71, 0x73]: return 'Error %s' % hex(v)

        if v == 0x73:
            return 'MPU9255 %s' % hex(v)
        elif v == 0x71:
            return 'MPU9250 %s' % hex(v)

    def WhoAmI_AK8963(self):
        '''
        Returns the ID fo magnetometer AK8963 if found.
        It should be 0x48.
        '''
        self.initMagnetometer()
        v = self.I2C.readBulk(self.AK8963_ADDRESS, 0, 1)[0]
        if v == 0x48:
            return 'AK8963 at %s' % hex(v)
        else:
            return 'AK8963 not found. returned :%s' % hex(v)

    def initMagnetometer(self):
        '''
        For MPU925x with integrated magnetometer.
        It's called a 10 DoF sensor, but technically speaking ,
        the 3-axis Accel , 3-Axis Gyro, temperature sensor are integrated in one IC, and the 3-axis magnetometer is implemented in a
        separate IC which can be accessed via an I2C passthrough.
        Therefore , in order to detect the magnetometer via an I2C scan, the passthrough must first be enabled on IC#1 (Accel,gyro,temp)
        '''
        self.I2C.writeBulk(self.ADDRESS, [self.INT_PIN_CFG, 0x22])  # I2C passthrough
        self.I2C.writeBulk(self.AK8963_ADDRESS, [self.AK8963_CNTL, 0])  # power down mag
        self.I2C.writeBulk(self.AK8963_ADDRESS,
                           [self.AK8963_CNTL, (1 << 4) | 6])  # mode   (0=14bits,1=16bits) <<4 | (2=8Hz , 6=100Hz)


if __name__ == "__main__":
    from PSL import sciencelab

    I = sciencelab.connect()
    A = connect(I.I2C)
    t, x, y, z = I.I2C.capture(A.ADDRESS, 0x43, 6, 5000, 1000, 'int')
    # print (t,x,y,z)
    from pylab import *

    plot(t, x)
    plot(t, y)
    plot(t, z)
    show()



"""Classes and functions related to the PSLab's oscilloscope instrument.

Example
-------
>>> from pslab import Oscilloscope
>>> scope = Oscilloscope()
>>> x, y1, y2, y3 = scope.capture(channels=3, samples=1600, timegap=2)
"""

import time
from typing import List, Tuple, Union

import numpy as np

import pslab.protocol as CP
from pslab.instrument.analog import ANALOG_CHANNELS, AnalogInput, GAIN_VALUES
from pslab.serial_handler import ADCBufferMixin, SerialHandler


class Oscilloscope(ADCBufferMixin):
    """Capture varying voltage signals on up to four channels simultaneously.

    Parameters
    ----------
    device : :class:`SerialHandler`, optional
        Serial interface for communicating with the PSLab device. If not
        provided, a new one will be created.
    """

    _CH234 = ["CH2", "CH3", "MIC"]

    def __init__(self, device: SerialHandler = None):
        self._device = SerialHandler() if device is None else device
        self._channels = {a: AnalogInput(a) for a in ANALOG_CHANNELS}
        self._channel_one_map = "CH1"
        self._trigger_voltage = None
        self._trigger_enabled = False
        self._trigger_channel = "CH1"
        self._set_gain("CH1", 1)
        self._set_gain("CH2", 1)

    def capture(
        self,
        channels: int,
        samples: int,
        timegap: float,
        trigger: Union[float, bool] = None,
        trigger_channel: str = None,
        block: bool = True,
    ) -> List[np.ndarray]:
        """Capture an oscilloscope trace from the specified input channels.

        Parameters
        ----------
        channels : str or {1, 2, 3, 4}
            Number of channels to sample from simultaneously, or the name
            (CH1, CH2, CH3, MIC, CAP, RES, VOL) of a single channel to sample
            from. If channel is an integer, the oscilloscope will sample the
            first one, two, three, or four channels in the aforementioned list.
        samples : int
            Number of samples to fetch. Maximum 10000 divided by number of
            channels.
        timegap : float
            Time gap between samples in microseconds. Will be rounded to the
            closest 1 / 8 µs. The minimum time gap depends on the type of
            measurement:

                +--------------+------------+----------+------------+
                | Simultaneous | No trigger | Trigger  | No trigger |
                | channels     | (10-bit)   | (10-bit) | (12-bit)   |
                +==============+============+==========+============+
                | 1            | 0.5 µs     | 0.75 µs  | 1 µs       |
                +--------------+------------+----------+------------+
                | 2            | 0.875 µs   | 0.875 µs | N/A        |
                +--------------+------------+----------+------------+
                | 4            | 1.75 µs    | 1.75 µs  | N/A        |
                +--------------+------------+----------+------------+

            Sample resolution is set automatically based on the above
            limitations; i.e. to get 12-bit samples only one channel may be
            sampled, there must be no active trigger, and the time gap must be
            1 µs or greater.
        trigger : float or bool, optional
            Voltage at which to trigger sampling. Triggering is disabled by
            default. Trigger settings persist between calls; disable by setting
            trigger=False.
        trigger_channel : str, optional
            Wait for the voltage level on this channel to cross the trigger
            value before sampling. Same as the first sampled channel by
            default.
        block : bool, optional
            Whether or not to block while sampling. If False, return timestamps
            immediately without waiting for corresponding voltages. User is
            responsible for waiting an appropriate amount of time before
            collecting samples with :meth:`fetch_data`. True by default.

        Example
        -------
        >>> from pslab import Oscilloscope
        >>> scope = Oscilloscope()
        >>> x, y = scope.capture(1, 3200, 1)

        Returns
        -------
        list of numpy.ndarray
            List of numpy arrays holding timestamps and corresponding voltages.
            In non-blocking mode, only timestamps are returned; voltages must
            be fetched using :meth:`fetch_data`.

        Raises
        ------
        ValueError
            If :channels: is not 1, 2, 3, 4, or one of CH1, CH2, CH3, MIC, CAP,
            RES, VOL, or
            :samples: > 10000 / :channels:, or
            :timegap: is too low.
        """
        if isinstance(channels, str):
            self._channel_one_map = channels
            channels = 1

        if trigger_channel is None:
            self._trigger_channel = self._channel_one_map
        else:
            self._trigger_channel = trigger_channel

        if trigger is False:
            self._trigger_enabled = False
        elif trigger is not None:
            if trigger != self._trigger_voltage:
                self.configure_trigger(voltage=trigger)

        self._check_args(channels, samples, timegap)
        timegap = int(timegap * 8) / 8
        self._capture(channels, samples, timegap)
        x = [timegap * np.arange(samples)]

        if block:
            time.sleep(1e-6 * samples * timegap)

            while not self.progress()[0]:
                pass

            # Discard MIC if user requested three channels.
            y = self.fetch_data()[:channels]

            return x + y
        else:
            return x

    def _check_args(self, channels: int, samples: int, timegap: float):
        if channels not in (1, 2, 3, 4):
            raise ValueError("Number of channels to sample must be 1, 2, 3, or 4.")

        max_samples = CP.MAX_SAMPLES // channels
        if not 0 < samples <= max_samples:
            e1 = f"Cannot collect more than {max_samples} when sampling from "
            e2 = f"{channels} channels."
            raise ValueError(e1 + e2)

        min_timegap = self._lookup_mininum_timegap(channels)
        if timegap < min_timegap:
            raise ValueError(f"timegap must be at least {min_timegap}.")

        if self._channel_one_map not in self._channels:
            e1 = f"{self._channel_one_map} is not a valid channel. "
            e2 = f"Valid channels are {list(self._channels.keys())}."
            raise ValueError(e1 + e2)

    def _lookup_mininum_timegap(self, channels: int) -> float:
        channels_idx = {
            1: 0,
            2: 1,
            4: 2,
        }
        min_timegaps = [[0.5, 0.75], [0.875, 0.875], [1.75, 1.75]]

        return min_timegaps[channels_idx[channels]][self.trigger_enabled]

    def _capture(self, channels: int, samples: int, timegap: float):
        self._invalidate_buffer()
        chosa = self._channels[self._channel_one_map].chosa
        self._channels[self._channel_one_map].resolution = 10
        self._device.send_byte(CP.ADC)

        CH123SA = 0  # TODO what is this?
        chosa = self._channels[self._channel_one_map].chosa
        self._channels[self._channel_one_map].samples_in_buffer = samples
        self._channels[self._channel_one_map].buffer_idx = 0
        if channels == 1:
            if self.trigger_enabled:
                self._device.send_byte(CP.CAPTURE_ONE)
                self._device.send_byte(chosa | 0x80)  # Trigger
            elif timegap >= 1:
                self._channels[self._channel_one_map].resolution = 12
                self._device.send_byte(CP.CAPTURE_DMASPEED)
                self._device.send_byte(chosa | 0x80)  # 12-bit mode
            else:
                self._device.send_byte(CP.CAPTURE_DMASPEED)
                self._device.send_byte(chosa)  # 10-bit mode
        elif channels == 2:
            self._channels["CH2"].resolution = 10
            self._channels["CH2"].samples_in_buffer = samples
            self._channels["CH2"].buffer_idx = 1 * samples
            self._device.send_byte(CP.CAPTURE_TWO)
            self._device.send_byte(chosa | (0x80 * self.trigger_enabled))
        else:
            for e, c in enumerate(self._CH234):
                self._channels[c].resolution = 10
                self._channels[c].samples_in_buffer = samples
                self._channels[c].buffer_idx = (e + 1) * samples
            self._device.send_byte(CP.CAPTURE_FOUR)
            self._device.send_byte(
                chosa | (CH123SA << 4) | (0x80 * self.trigger_enabled)
            )

        self._device.send_int(samples)
        self._device.send_int(int(timegap * 8))  # 8 MHz clock
        self._device.get_ack()

    def _invalidate_buffer(self):
        for c in self._channels.values():
            c.samples_in_buffer = 0
            c.buffer_idx = None

    def fetch_data(self) -> List[np.ndarray]:
        """Fetch captured samples.

        Example
        -------
        >>> from pslab import Oscilloscope
        >>> scope = Oscilloscope()
        >>> scope.capture_nonblocking(channels=2, samples=1600, timegap=1)
        >>> y1, y2 = scope.fetch_data()

        Returns
        -------
        list of numpy.ndarray
            List of numpy arrays holding sampled voltages.
        """
        channels = [c for c in self._channels.values() if c.samples_in_buffer]
        data = [None] * len(channels)

        for i, channel in enumerate(channels):
            samples = channel.samples_in_buffer
            data[i] = self.fetch_buffer(samples, channel.buffer_idx)
            data[i] = channel.scale(np.array(data[i]))

        return data

    def progress(self) -> Tuple[bool, int]:
        """Return the status of a capture call.

        Returns
        -------
        bool, int
            A boolean indicating whether the capture is complete, followed by
            the number of samples currently held in the buffer.
        """
        self._device.send_byte(CP.ADC)
        self._device.send_byte(CP.GET_CAPTURE_STATUS)
        conversion_done = self._device.get_byte()
        samples = self._device.get_int()
        self._device.get_ack()

        return bool(conversion_done), samples

    def configure_trigger(
        self,
        channel: str = None,
        voltage: float = 0,
        prescaler: int = 0,
        enable: bool = True,
    ):
        """Configure trigger parameters for 10-bit capture routines.

        The capture routines will wait until a rising edge of the input signal
        crosses the specified level. The trigger will timeout within 8 ms, and
        capture will start regardless.

        To disable the trigger after configuration, set the trigger_enabled
        attribute of the Oscilloscope instance to False.

        Parameters
        ----------
        channel : {'CH1', 'CH2', 'CH3', 'MIC', 'CAP', 'RES', 'VOL'}, optional
            The name of the trigger channel. First sampled channel by default.
        voltage : float, optional
            The trigger voltage in volts. The default value is 0.
        prescaler : int, optional
            The default value is 0.
        enable_trigger : bool, optional
            Set this to False to disable the trigger. True by default.

        Examples
        --------
        >>> from pslab import Oscilloscope
        >>> scope = Oscilloscope()
        >>> scope.configure_trigger(channel='CH1', voltage=1.1)
        >>> xy = scope.capture(channels=1, samples=800, timegap=2)
        >>> diff = abs(xy[1, 0] - 1.1)  # Should be small unless a timeout occurred.

        Raises
        ------
        TypeError
            If the trigger channel is set to a channel which cannot be sampled.
        """
        if enable is False:
            self._trigger_enabled = False
            return

        if channel is not None:
            self._trigger_channel = channel

        if self.trigger_channel == self._channel_one_map:
            channel = 0
        elif self.trigger_channel in self._CH234:
            channel = self._CH234.index(self.trigger_channel) + 1
        else:
            raise TypeError(f"Cannot trigger on {self.trigger_channel}.")

        self._device.send_byte(CP.ADC)
        self._device.send_byte(CP.CONFIGURE_TRIGGER)
        # Trigger channel (4lsb) , trigger timeout prescaler (4msb)
        self._device.send_byte((prescaler << 4) | (1 << channel))  # TODO prescaler?
        level = self._channels[self.trigger_channel].unscale(voltage)
        self._device.send_int(level)
        self._device.get_ack()
        self._trigger_enabled = True

    @property
    def trigger_enabled(self) -> bool:
        """bool: Wait for trigger condition before capture start."""
        return self._trigger_enabled

    @property
    def trigger_channel(self) -> str:
        """str: Name of channel to trigger on."""
        return self._trigger_channel

    @property
    def trigger_voltage(self) -> float:
        """float: Trigger when voltage crosses this value."""
        return self._trigger_voltage

    def select_range(self, channel: str, voltage_range: Union[int, float]):
        """Set appropriate gain automatically.

        Setting the right voltage range will result in better resolution.

        Parameters
        ----------
        channel : {'CH1', 'CH2'}
            Channel on which to apply gain.
        voltage_range : {16, 8, 4, 3, 2, 1.5, 1, .5}

        Examples
        --------
        Set 2x gain on CH1. Voltage range ±8 V:

        >>> from pslab import Oscilloscope
        >>> scope = Oscilloscope()
        >>> scope.select_range('CH1', 8)
        """
        ranges = [16, 8, 4, 3, 2, 1.5, 1, 0.5]
        gain = GAIN_VALUES[ranges.index(voltage_range)]
        self._set_gain(channel, gain)

    def _set_gain(self, channel: str, gain: int):
        self._channels[channel].gain = gain
        pga = self._channels[channel].programmable_gain_amplifier
        gain_idx = GAIN_VALUES.index(gain)
        self._device.send_byte(CP.ADC)
        self._device.send_byte(CP.SET_PGA_GAIN)
        self._device.send_byte(pga)
        self._device.send_byte(gain_idx)
        self._device.get_ack()



"""The PSLab's multimeter can measure voltage, resistance, and capacitance."""
import time
from typing import Tuple

import numpy as np
from scipy.optimize import curve_fit

import pslab.protocol as CP
from pslab.instrument.analog import GAIN_VALUES, INPUT_RANGES
from pslab.instrument.oscilloscope import Oscilloscope
from pslab.serial_handler import SerialHandler

_MICROSECONDS = 1e-6


class Multimeter(Oscilloscope):
    """Measure voltage, resistance and capacitance.

    Parameters
    ----------
    device : Handler
        Serial interface for communicating with the PSLab device. If not
        provided, a new one will be created.
    """

    _CURRENTS = [5.5e-4, 5.5e-7, 5.5e-6, 5.5e-5]
    _CURRENTS_RANGES = [1, 2, 3, 0]  # Smallest first,
    _RC_RESISTANCE = 1e4
    _CAPACITOR_CHARGED_VOLTAGE = 0.9 * max(INPUT_RANGES["CAP"])
    _CAPACITOR_DISCHARGED_VOLTAGE = 0.01 * max(INPUT_RANGES["CAP"])

    def __init__(self, device: SerialHandler = None):
        self._stray_capacitance = 5e-11
        super().__init__(device)

    def measure_resistance(self) -> float:
        """Measure the resistance of a resistor connected between RES and GND.

        Returns
        -------
        resistance : float
            Resistance in ohm.
        """
        voltage = self.measure_voltage("RES")
        resolution = max(INPUT_RANGES["RES"]) / (
            2 ** self._channels["RES"].resolution - 1
        )

        if voltage >= max(INPUT_RANGES["RES"]) - resolution:
            return np.inf

        pull_up_resistance = 5.1e3
        current = (INPUT_RANGES["RES"][1] - voltage) / pull_up_resistance
        return voltage / current

    def measure_voltage(self, channel: str = "VOL") -> float:
        """Measure the voltage on the selected channel.

        Parameters
        ----------
        channel : {"CH1", "CH2", "CH3", "MIC", "CAP", "RES", "VOL"}, optional
            The name of the analog input on which to measure the voltage. The
            default channel is VOL.

        Returns
        -------
        voltage : float
            Voltage in volts.
        """
        self._voltmeter_autorange(channel)
        return self._measure_voltage(channel)

    def _measure_voltage(self, channel: str) -> float:
        self._channels[channel].resolution = 12
        scale = self._channels[channel].scale
        chosa = self._channels[channel].chosa
        self._device.send_byte(CP.ADC)
        self._device.send_byte(CP.GET_VOLTAGE_SUMMED)
        self._device.send_byte(chosa)
        raw_voltage_sum = self._device.get_int()  # Sum of 16 samples.
        self._device.get_ack()
        raw_voltage_mean = round(raw_voltage_sum / 16)
        voltage = scale(raw_voltage_mean)
        return voltage

    def _voltmeter_autorange(self, channel: str) -> float:
        if channel in ("CH1", "CH2"):
            self._set_gain(channel, 1)  # Reset gain.
            voltage = self._measure_voltage(channel)

            for gain in GAIN_VALUES[::-1]:
                rng = max(INPUT_RANGES[channel]) / gain
                if abs(voltage) < rng:
                    break

            self._set_gain(channel, gain)

            return rng
        else:
            return max(INPUT_RANGES[channel])

    def calibrate_capacitance(self):
        """Calibrate stray capacitance.

        Correctly calibrated stray capacitance is important when measuring
        small capacitors (picofarad range).

        Stray capacitace should be recalibrated if external wiring is connected
        to the CAP pin.
        """
        for charge_time in np.unique(np.int16(np.logspace(2, 3))):
            self._discharge_capacitor()
            voltage, capacitance = self._measure_capacitance(1, 0, charge_time)
            if voltage >= self._CAPACITOR_CHARGED_VOLTAGE:
                break
        self._stray_capacitance += capacitance

    def measure_capacitance(self) -> float:
        """Measure the capacitance of a capacitor connected between CAP and GND.

        Returns
        -------
        capacitance : float
            Capacitance in Farad.
        """
        for current_range in self._CURRENTS_RANGES:
            for i, charge_time in enumerate([50000, 5000, 500, 50, 5]):
                voltage, _ = self._measure_capacitance(current_range, 0, charge_time)

                if voltage < self._CAPACITOR_CHARGED_VOLTAGE:
                    if i:
                        return self._binary_search_capacitance(
                            current_range, charge_time, charge_time * 10
                        )
                    else:
                        break  # Increase current.

        # Capacitor too big, use alternative method.
        return self._measure_rc_capacitance()

    def _binary_search_capacitance(
        self,
        current_range: int,
        low_charge_time: int,
        high_charge_time: int,
    ) -> float:
        charge_time = (high_charge_time + low_charge_time) // 2
        voltage, capacitance = self._measure_capacitance(
            current_range,
            0,
            charge_time,
        )

        if voltage / self._CAPACITOR_CHARGED_VOLTAGE < 0.98:
            return self._binary_search_capacitance(
                current_range,
                charge_time,
                high_charge_time,
            )
        elif voltage / self._CAPACITOR_CHARGED_VOLTAGE > 1.02:
            return self._binary_search_capacitance(
                current_range,
                low_charge_time,
                charge_time,
            )
        else:
            return capacitance

    def _set_cap(self, state, charge_time):
        """Set CAP HIGH or LOW."""
        self._device.send_byte(CP.ADC)
        self._device.send_byte(CP.SET_CAP)
        self._device.send_byte(state)
        self._device.send_int(charge_time)
        self._device.get_ack()

    def _discharge_capacitor(
        self, discharge_time: int = 50000, timeout: float = 1
    ) -> float:
        start_time = time.time()
        voltage = previous_voltage = self.measure_voltage("CAP")

        while voltage > self._CAPACITOR_DISCHARGED_VOLTAGE:
            self._set_cap(0, discharge_time)
            voltage = self.measure_voltage("CAP")

            if abs(previous_voltage - voltage) < self._CAPACITOR_DISCHARGED_VOLTAGE:
                break

            previous_voltage = voltage

            if time.time() - start_time > timeout:
                break

        return voltage

    def _measure_capacitance(
        self, current_range: int, trim: int, charge_time: int
    ) -> Tuple[float, float]:
        self._discharge_capacitor()
        self._channels["CAP"].resolution = 12
        self._device.send_byte(CP.COMMON)
        self._device.send_byte(CP.GET_CAPACITANCE)
        self._device.send_byte(current_range)

        if trim < 0:
            self._device.send_byte(int(31 - abs(trim) / 2) | 32)
        else:
            self._device.send_byte(int(trim / 2))

        self._device.send_int(charge_time)
        time.sleep(charge_time * _MICROSECONDS)
        raw_voltage = self._device.get_int()
        voltage = self._channels["CAP"].scale(raw_voltage)
        self._device.get_ack()
        charge_current = self._CURRENTS[current_range] * (100 + trim) / 100

        if voltage:
            capacitance = (
                charge_current * charge_time * _MICROSECONDS / voltage
                - self._stray_capacitance
            )
        else:
            capacitance = 0

        return voltage, capacitance

    def _measure_rc_capacitance(self) -> float:
        """Measure the capacitance by discharge through a 10K resistor."""
        (x,) = self.capture("CAP", CP.MAX_SAMPLES, 10, block=False)
        x *= _MICROSECONDS
        self._set_cap(1, 50000)  # charge
        self._set_cap(0, 50000)  # discharge
        (y,) = self.fetch_data()

        if y.max() >= self._CAPACITOR_CHARGED_VOLTAGE:
            discharge_start = np.where(y >= self._CAPACITOR_CHARGED_VOLTAGE)[0][-1]
        else:
            discharge_start = np.where(y == y.max())[0][-1]

        x = x[discharge_start:]
        y = y[discharge_start:]

        # CAP floats for a brief period of time (~500 µs) between being set
        # HIGH until it is set LOW. This data is not useful and should be
        # discarded. When CAP is set LOW the voltage declines sharply, which
        # manifests as a negative peak in the time derivative.
        dydx = np.diff(y) / np.diff(x)
        cap_low = np.where(dydx == dydx.min())[0][0]
        x = x[cap_low:]
        y = y[cap_low:]

        # Discard data after the voltage reaches zero (improves fit).
        try:
            v_zero = np.where(y == 0)[0][0]
            x = x[:v_zero]
            y = y[:v_zero]
        except IndexError:
            pass

        # Remove time offset.
        x -= x[0]

        def discharging_capacitor_voltage(
            x: np.ndarray, v_init: float, rc_time_constant: float
        ) -> np.ndarray:
            return v_init * np.exp(-x / rc_time_constant)

        # Solve discharging_capacitor_voltage for rc_time_constant.
        rc_time_constant_guess = (-x[1:] / np.log(y[1:] / y[0])).mean()
        guess = [y[0], rc_time_constant_guess]
        popt, _ = curve_fit(discharging_capacitor_voltage, x, y, guess)
        rc_time_constant = popt[1]
        rc_capacitance = rc_time_constant / self._RC_RESISTANCE
        return rc_capacitance
